// This is beautiful, essentially every primitive is generated by a single function!
#include "ggen.h"

using namespace henge;

static bool generic_heightfield(trimesh *mesh, int usub, int vsub,
		float umax, float vmax,	gen_sample_func dist_func, void *dist_func_cls,
		Vector3 (*vertex)(float, float, float), tex_coord_gen tcgen);

static Vector3 vertex_revol(float u, float v, float dist);
static Vector3 vertex_plane(float u, float v, float dist);

static Vector2 tex_xz(float u, float v, const Vector3 &vert);
static Vector2 tex_xy(float u, float v, const Vector3 &vert);
static Vector2 tex_uv(float u, float v, const Vector3 &vert);


// this array must remain in sync with the tex_coord_gen enumeration in ggen.h

static Vector2 (*tex_func[])(float, float, const Vector3&) = {
	tex_xz,
	tex_xy,
	tex_uv
};

/* gen_revol (surface of revolution) is a special case of a heightfield, defined
 * by the distance around the Y axis.
 */
bool henge::gen_revol(trimesh *mesh, int usub, int vsub, float umax, float vmax,
		gen_sample_func dist_func, void *dist_func_cls, tex_coord_gen tcgen)
{
	return generic_heightfield(mesh, usub, vsub, umax, vmax, dist_func, dist_func_cls,
			vertex_revol, tcgen);
}

/* gen_height (heighfield) is also defined in terms of the generic_heightfield
 * function
 */
bool henge::gen_height(trimesh *mesh, int usub, int vsub, gen_sample_func height_func,
		void *height_func_cls, tex_coord_gen tcgen)
{
	return generic_heightfield(mesh, usub, vsub, 1, 1, height_func, height_func_cls,
			vertex_plane, tcgen);
}

/* Surface of revolution distance evaluator for a sphere.
 * The closure pointer points to the radius of the cylinder.
 */
static float sph_dist_func(float u, float v, void *cls)
{
	return fabs(sin(v * PI));
}

// special case because we want the sphere to have radius, not diameter of 1
static Vector3 vertex_revol_sph(float u, float v, float dist)
{
	return Vector3(sin(u * TWO_PI) * dist, 2.0 * (v - 0.5), cos(u * TWO_PI) * dist).normalized();
}

// a sphere is a surface of revolution ... 
bool henge::gen_sphere(trimesh *mesh, float rad, int slices, int stacks, float umax,
		float vmax, tex_coord_gen tcgen)
{
	if(!generic_heightfield(mesh, slices, stacks, umax, vmax, sph_dist_func, &rad,
			vertex_revol_sph, tcgen)) {
		return false;
	}

	Vector3 *vert = mesh->get_data_vec3(EL_VERTEX);
	Vector3 *norm = mesh->get_data_vec3(EL_NORMAL);
	for(int i=0; i<mesh->get_count(EL_VERTEX); i++) {
		norm[i] = vert[i].normalized();	// recalc normals to avoid singularity at the poles
		vert[i] *= rad;	// this is where we make the sphere the proper size
	}
	return true;
}

/* Surface of revolution distance evaluator for a cylinder.
 * The closure pointer points to the radius of the cylinder.
 */
static float cyl_dist_func(float u, float v, void *cls)
{
	return 0.5f * *(float*)cls;
}

// a cylinder is a surface of revolution ...
bool henge::gen_cylinder(trimesh *mesh, float rad, int slices, int stacks, float umax,
		tex_coord_gen tcgen)
{
	return gen_revol(mesh, slices, stacks, umax, 1, cyl_dist_func, &rad, tcgen);
}


static float plane_dist_func(float u, float v, void *cls)
{
	return 0.0f;
}

// a plane is a heightfield with a constant height of 0
bool henge::gen_plane(trimesh *mesh, int usub, int vsub, tex_coord_gen tcgen)
{
	return gen_height(mesh, usub, vsub, plane_dist_func, 0, tcgen);
}


/* The following static functions are either helpers (see tex_* tang_*) or
 * common code of the gen_* functions above that has been factored out
 * (see generic_heightfield).
 */

#define QUAD(a, b, c, d)	\
	do {	\
		*iptr++ = a;	\
		*iptr++ = b;	\
		*iptr++ = c;	\
		*iptr++ = a;	\
		*iptr++ = c;	\
		*iptr++ = d;	\
	} while(0)

static Vector3 vertex_revol(float u, float v, float dist)
{
	return Vector3(sin(u * TWO_PI) * dist, v - 0.5, cos(u * TWO_PI) * dist);
}

static Vector3 vertex_plane(float u, float v, float dist)
{
	return Vector3(2.0 * u - 1.0, dist, 2.0 * (0.5 - v));
}

/* right... this function can generate both heightfields on a plane, and around
 * an axis (surface of revolution), by just supplying the correct vertex(u, v, dist)
 * function (see above). So it's the common base for both gen_revol and gen_height.
 */
static bool generic_heightfield(trimesh *mesh, int usub, int vsub,
		float umax, float vmax,	gen_sample_func dist_func, void *dist_func_cls,
		Vector3 (*vertex)(float, float, float), tex_coord_gen tcgen)
{
	if(!mesh || !dist_func) {
		return false;
	}

	int unum = usub + 1;
	int vnum = vsub + 1;
	int num_verts = unum * vnum;
	int num_quads = usub * vsub;
	int num_tris = num_quads * 2;
	int num_idx = num_tris * 3;

	// allocate mesh data storage
	if(!mesh->set_data(EL_VERTEX, (Vector3*)0, num_verts) ||
			!mesh->set_data(EL_NORMAL, (Vector3*)0, num_verts) ||
			!mesh->set_data(EL_TANGENT, (Vector3*)0, num_verts) ||
			!mesh->set_data(EL_TEXCOORD, (Vector2*)0, num_verts) ||
			!mesh->set_data(EL_INDEX, (unsigned int*)0, num_idx)) {
		return false;
	}

	// retrieve pointers to the allocated arrays
	Vector3 *vptr = mesh->get_data_vec3(EL_VERTEX);
	Vector3 *nptr = mesh->get_data_vec3(EL_NORMAL);
	Vector3 *tptr = mesh->get_data_vec3(EL_TANGENT);
	Vector2 *xptr = mesh->get_data_vec2(EL_TEXCOORD);
	unsigned int *iptr = mesh->get_data_int(EL_INDEX);

	// generate vertices
	for(int i=0; i<unum; i++) {
		float u = umax * (float)i / (float)usub;

		for(int j=0; j<vnum; j++) {
			float v = vmax * (float)j / (float)vsub;

			float dist = dist_func(u, v, dist_func_cls);
			Vector3 vert = vertex(u, v, dist);

			Vector3 uprev = vertex(u - 0.01, v, dist_func(u - 0.01, v, dist_func_cls));
			Vector3 unext = vertex(u + 0.01, v, dist_func(u + 0.01, v, dist_func_cls));
			Vector3 tang = unext - uprev;

			Vector3 vprev = vertex(u, v - 0.01, dist_func(u, v - 0.01, dist_func_cls));
			Vector3 vnext = vertex(u, v + 0.01, dist_func(u, v + 0.01, dist_func_cls));
			Vector3 bitan = vnext - vprev;

			Vector3 norm = cross_product(tang, bitan).normalized();

			*vptr++ = vert;
			*nptr++ = norm;
			*tptr++ = tang.normalized();
			*xptr++ = tex_func[tcgen](u, v, vert);
		}
	}

	// construct indices
	for(int i=0; i<usub; i++) {
		int base = i * vnum;
		for(int j=0; j<vsub; j++) {
			QUAD(j + base,
				vnum + j + base,
				vnum + 1 + j + base,
				1 + j + base);
		}
	}

	return true;
}

static Vector2 tex_xz(float u, float v, const Vector3 &vert)
{
	return Vector2(vert.x * 0.5 + 0.5, vert.z * 0.5 + 0.5);
}

static Vector2 tex_xy(float u, float v, const Vector3 &vert)
{
	return Vector2(vert.x * 0.5 + 0.5, vert.y * 0.5 + 0.5);
}

static Vector2 tex_uv(float u, float v, const Vector3 &vert)
{
	return Vector2(u, v);
}
