#include <assert.h>
#include <float.h>
#include "opengl.h"
#include "mesh.h"
#include "sdr.h"
#include "kdtree.h"
#include "errlog.h"

/* Define DEBUG_DRAWING to force drawing of meshes through immediate mode calls.
 * Sometimes it's helpful to see index indirections etc happen in the cpu directly
 * easier to spot&debug fucked up index buffers, missing attribute buffers, etc.
 */
#undef DEBUG_DRAWING

using namespace henge;

trimesh::trimesh()
{
	init();
}

trimesh::trimesh(const trimesh &m)
{
	init();
	*this = m;
}

void trimesh::init()
{
	vert = norm = tang = 0;
	tc = 0;
	index = 0;
	nvert = nnorm = ntang = ntc = nindex = 0;

	dynamic = false;
	dlist = 0;
	
	if(caps.vbo) {
		glGenBuffersARB(EL_COUNT, vbo);
		memset(vbo_valid, 0, EL_COUNT * sizeof *vbo_valid);
	}

	kdt_valid = false;
	bounds_valid = false;
}

trimesh &trimesh::operator =(const trimesh &m)
{
	if(this == &m) {
		return *this;
	}
	nvert = m.nvert;
	nnorm = m.nnorm;
	ntang = m.ntang;
	ntc = m.ntc;
	nindex = m.nindex;

	dynamic = m.dynamic;

	// copy the various arrays
	if(m.vert) {
		vert = new Vector3[nvert];
		memcpy(vert, m.vert, nvert * sizeof *vert);
	}
	if(m.norm) {
		norm = new Vector3[nnorm];
		memcpy(norm, m.norm, nnorm * sizeof *norm);
	}
	if(m.tang) {
		tang = new Vector3[ntang];
		memcpy(tang, m.tang, ntang * sizeof *tang);
	}
	if(m.tc) {
		tc = new Vector2[ntc];
		memcpy(tc, m.tc, ntc * sizeof *tc);
	}
	if(m.index) {
		index = new unsigned int[nindex];
		memcpy(index, m.index, nindex * sizeof *index);
	}

	return *this;
}

trimesh::~trimesh()
{
	delete [] vert;
	delete [] norm;
	delete [] tang;
	delete [] tc;
	delete [] index;

	if(dlist) {
		glDeleteLists(dlist, 1);
	}

	if(vbo[0]) {
		glDeleteBuffersARB(EL_COUNT, vbo);
	}
}


bool trimesh::merge(const trimesh &mesh)
{
	int vidx_offs = 0;

	if((index && !mesh.index) || (!index && mesh.index)) {
		error("merging of an indexed with a non-indexed mesh not supported yet");
		return false;
	}

	// merge positions
	if(vert || mesh.vert) {
		Vector3 *new_vert = new Vector3[nvert + mesh.nvert];
		memcpy(new_vert, vert, nvert * sizeof *new_vert);
		memcpy(new_vert + nvert, mesh.vert, mesh.nvert * sizeof *new_vert);
		vidx_offs = nvert;

		delete [] vert;
		vert = new_vert;
		nvert += mesh.nvert;
		kdt_valid = false;
		bounds_valid = false;

		vbo_valid[EL_VERTEX] = false;
	}

	// merge normals
	if(norm || mesh.norm) {
		Vector3 *new_norm = new Vector3[nnorm + mesh.nnorm];
		memcpy(new_norm, norm, nnorm * sizeof *new_norm);
		memcpy(new_norm + nnorm, mesh.norm, mesh.nnorm * sizeof *new_norm);

		delete [] norm;
		norm = new_norm;
		nnorm += mesh.nnorm;
		vbo_valid[EL_NORMAL] = false;
	}

	// merge tangents
	if(tang || mesh.tang) {
		Vector3 *new_tang = new Vector3[ntang + mesh.ntang];
		memcpy(new_tang, tang, ntang * sizeof *new_tang);
		memcpy(new_tang + ntang, mesh.tang, mesh.ntang * sizeof *new_tang);

		delete [] tang;
		tang = new_tang;
		ntang += mesh.ntang;
		vbo_valid[EL_TANGENT] = false;
	}

	// merge tex coords
	if(tc || mesh.tc) {
		Vector2 *new_tc = new Vector2[ntc + mesh.ntc];
		memcpy(new_tc, tc, ntc * sizeof *new_tc);
		memcpy(new_tc + ntc, mesh.tc, mesh.ntc * sizeof *new_tc);

		delete [] tc;
		tc = new_tc;
		ntc += mesh.ntc;
		vbo_valid[EL_TEXCOORD] = false;
	}

	// merge indices
	if(index && mesh.index) {
		unsigned int *new_index = new unsigned int[nindex + mesh.nindex];
		memcpy(new_index, index, nindex * sizeof *new_index);

		unsigned int *iptr = new_index + nindex;
		for(int i=0; i<mesh.nindex; i++) {
			*iptr++ = mesh.index[i] + vidx_offs;
		}

		delete [] index;
		index = new_index;
		nindex += mesh.nindex;
		vbo_valid[EL_INDEX] = false;
	}

	if(dlist) {
		glDeleteLists(1, dlist);
		dlist = 0;
	}
	return true;
}


bool trimesh::set_data(int elem, const Vector3 *data, int count)
{
	Vector3 *varr;
	try {
		varr = new Vector3[count];
	}
	catch(...) {
		return false;
	}

	switch(elem) {
	case EL_VERTEX:
		delete [] vert;
		vert = varr;
		nvert = count;
		kdt_valid = false;
		bounds_valid = false;
		break;

	case EL_NORMAL:
		delete [] norm;
		norm = varr;
		nnorm = count;
		break;

	case EL_TANGENT:
		delete [] tang;
		tang = varr;
		ntang = count;
		break;

	default:
		delete [] varr;
		return false;
	}
	
	if(data) {
		memcpy(varr, data, count * sizeof *varr);
	}

	vbo_valid[elem] = false;
	if(dlist) {
		glDeleteLists(1, dlist);
		dlist = 0;
	}

	return true;
}

bool trimesh::set_data(int elem, const Vector2 *data, int count)
{
	if(elem != EL_TEXCOORD) {
		return false;
	}

	try {
		Vector2 *new_tc = new Vector2[count];
		delete [] tc;
		tc = new_tc;
	}
	catch(...) {
		return false;
	}

	if(data) {
		memcpy(tc, data, count * sizeof *tc);
	}
	ntc = count;

	vbo_valid[elem] = false;
	if(dlist) {
		glDeleteLists(1, dlist);
		dlist = 0;
	}

	return true;
}

bool trimesh::set_data(int elem, const unsigned int *data, int count)
{
	if(elem != EL_INDEX) {
		return false;
	}

	try {
		unsigned int *idx = new unsigned int[count];
		delete [] index;
		index = idx;
	}
	catch(...) {
		return false;
	}

	if(data) {
		memcpy(index, data, count * sizeof *index);
	}
	nindex = count;

	vbo_valid[elem] = false;
	if(dlist) {
		glDeleteLists(1, dlist);
		dlist = 0;
	}

	kdt_valid = false;

	return true;
}

Vector3 *trimesh::get_data_vec3(int elem)
{
	bool prev = vbo_valid[elem];
	vbo_valid[elem] = false;

	switch(elem) {
	case EL_VERTEX:
		kdt_valid = false;
		return vert;

	case EL_NORMAL:
		return norm;

	case EL_TANGENT:
		return tang;

	default:
		vbo_valid[elem] = prev;
	}
	return 0;
}

const Vector3 *trimesh::get_data_vec3(int elem) const
{
	switch(elem) {
	case EL_VERTEX:
		return vert;

	case EL_NORMAL:
		return norm;

	case EL_TANGENT:
		return tang;

	default:
		break;
	}
	return 0;
}

Vector2 *trimesh::get_data_vec2(int elem)
{
	if(elem == EL_TEXCOORD) {
		vbo_valid[elem] = false;
		return tc;
	}
	return 0;
}

const Vector2 *trimesh::get_data_vec2(int elem) const
{
	if(elem == EL_TEXCOORD) {
		return tc;
	}
	return 0;
}

unsigned int *trimesh::get_data_int(int elem)
{
	if(elem == EL_INDEX) {
		vbo_valid[elem] = false;
		kdt_valid = false;
		return index;
	}
	return 0;
}

const unsigned int *trimesh::get_data_int(int elem) const
{
	if(elem == EL_INDEX) {
		return index;
	}
	return 0;
}

int trimesh::get_count(int elem) const
{
	switch(elem) {
	case EL_VERTEX:
		return nvert;

	case EL_NORMAL:
		return nnorm;

	case EL_TANGENT:
		return ntang;

	case EL_TEXCOORD:
		return ntc;

	case EL_INDEX:
		return nindex;
	}
	return 0;
}


void trimesh::build_kdtree()
{
	kdt.clear();

	for(int i=0; i<nvert; i++) {
		// XXX last argument is triangle index... won't work for indexed triangles
		kdt.insert(vert[i].x, vert[i].y, vert[i].z, i / 3);
	}
}

/* TODO move the adjacency information outside of this function
 * to be calculated whenever the arrays change.
 *
 * also implement this correctly for the case of indexed triangles.
 */
void trimesh::calc_normals()
{
	Vector3 *tri_norm = 0;
	int **adj_tri = 0;
	int *adj_tri_num = 0;

	if(!kdt_valid) {
		build_kdtree();
		kdt_valid = true;
	}

	int tcount = nvert / 3;
	try {
		tri_norm = new Vector3[tcount];
		adj_tri = new int*[nvert];
		adj_tri_num = new int[nvert];
	}
	catch(...) {
		delete [] tri_norm;
		delete [] adj_tri;
		return;
	}

	// calculate triangle normals
	for(int i=0; i<tcount; i++) {
		Vector3 v[3];
		v[0] = vert[i * 3];
		v[1] = vert[i * 3 + 1];
		v[2] = vert[i * 3 + 2];

		Vector3 a = v[1] - v[0];
		Vector3 b = v[2] - v[0];
		tri_norm[i] = cross_product(a, b);

		// for every vertex of the triangle
		for(int j=0; j<3; j++) {

			// find its neighboring vertices
			kdres<int> *res = kdt.nearest(v[j].x, v[j].y, v[j].z, 0.0001f);
			int res_size = res->size();

			int vidx = i * 3 + j;
			adj_tri_num[vidx] = res_size;
			adj_tri[vidx] = new int[res_size];

			for(int k=0; k<res_size; k++) {
				adj_tri[vidx][k] = res->get_data() / 3;
				res->next();
			}
			delete res;
		}
	}

	for(int i=0; i<nvert; i++) {
		Vector3 n;
		for(int j=0; j<adj_tri_num[i]; j++) {
			n += tri_norm[adj_tri[i][j]];
		}
		n.normalize();
		norm[i] = n.normalized();	// XXX what if nvert != nnorm ?

		delete [] adj_tri[i];
	}

	delete [] adj_tri;
	delete [] adj_tri_num;
	delete [] tri_norm;
}


#define SWAP(type, a, b)	\
	do {					\
		type tmp = a;		\
		a = b;				\
		b = tmp;			\
	} while(0)


void trimesh::flip_winding()
{
	if(index) {
		for(int i=0; i<nindex; i++) {
			if(i % 3 == 2) {
				SWAP(int, index[i], index[i - 2]);
			}
		}
		// invalidate index buffer object
		vbo_valid[EL_INDEX] = false;
	} else {
		for(int i=0; i<nvert; i++) {
			if(i % 3 == 2) {
				if(vert) {
					SWAP(Vector3, vert[i], vert[i - 2]);
				}
				if(norm) {
					assert(nvert == nnorm);
					SWAP(Vector3, norm[i], norm[i - 2]);
				}
				if(tang) {
					assert(nvert == ntang);
					SWAP(Vector3, tang[i], tang[i - 2]);
				}
				if(tc) {
					assert(nvert == ntc);
					SWAP(Vector2, tc[i], tc[i - 2]);
				}
			}
		}
		// invalidate all changed vbos
		vbo_valid[EL_VERTEX] = false;
		vbo_valid[EL_NORMAL] = false;
		vbo_valid[EL_TANGENT] = false;
		vbo_valid[EL_TEXCOORD] = false;
	}

	// invalidate the display list
	if(dlist) {
		glDeleteLists(dlist, 1);
		dlist = 0;
	}
}

void trimesh::flip_normals()
{
	for(int i=0; i<nvert; i++) {
		norm[i] = -norm[i];
	}
	vbo_valid[EL_NORMAL] = false;

	// invalidate the display list
	if(dlist) {
		glDeleteLists(dlist, 1);
		dlist = 0;
	}
}

void trimesh::transform(const Matrix4x4 &mat)
{
	Matrix4x4 norm_mat;

	for(int i=0; i<nvert; i++) {
		vert[i].transform(mat);
	}
	vbo_valid[EL_VERTEX] = false;
	bounds_valid = false;

	if(norm || tang) {
		norm_mat = mat.inverse().transposed();
	}

	if(norm) {
		for(int i=0; i<nnorm; i++) {
			norm[i].transform(norm_mat);
		}
		vbo_valid[EL_NORMAL] = false;
	}
	if(tang) {
		for(int i=0; i<ntang; i++) {
			tang[i].transform(norm_mat);
		}
		vbo_valid[EL_TANGENT] = false;
	}

	// invalidate the display list
	if(dlist) {
		glDeleteLists(dlist, 1);
		dlist = 0;
	}
}

Vector3 trimesh::get_centroid() const
{
	if(!bounds_valid) {
		((trimesh*)this)->calc_bounds();
	}
	return centroid;
}

Vector3 trimesh::get_aabb_min() const
{
	if(!bounds_valid) {
		((trimesh*)this)->calc_bounds();
	}
	return aabb_min;
}

Vector3 trimesh::get_aabb_max() const
{
	if(!bounds_valid) {
		((trimesh*)this)->calc_bounds();
	}
	return aabb_max;
}

/* TODO optimize updates, use glBufferSubDataARB instead of glBufferDataARB */
void trimesh::setup_vertex_arrays() const
{
	// do we have vertices? (we better have some...)
	if(vert) {
		glEnableClientState(GL_VERTEX_ARRAY);

		if(caps.vbo) {
			glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo[EL_VERTEX]);

			// if it's invalid, update it
			if(!vbo_valid[EL_VERTEX]) {
				glBufferDataARB(GL_ARRAY_BUFFER_ARB, nvert * sizeof *vert, vert, GL_DYNAMIC_DRAW_ARB);
				vbo_valid[EL_VERTEX] = true;
			}
			glVertexPointer(3, GL_FLOAT, 0, 0);
		} else {
			glVertexPointer(3, GL_FLOAT, 0, vert);
		}
	}

	// do we have normals?
	if(norm) {
		glEnableClientState(GL_NORMAL_ARRAY);

		if(caps.vbo) {
			glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo[EL_NORMAL]);

			// if it's invalid, update it
			if(!vbo_valid[EL_NORMAL]) {
				glBufferDataARB(GL_ARRAY_BUFFER_ARB, nnorm * sizeof *norm, norm, GL_DYNAMIC_DRAW_ARB);
				vbo_valid[EL_NORMAL] = true;
			}
			glNormalPointer(GL_FLOAT, 0, 0);
		} else {
			glNormalPointer(GL_FLOAT, 0, norm);
		}
	}

	// do we have texture coordinates?
	if(tc) {
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		if(caps.vbo) {
			glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo[EL_TEXCOORD]);

			// if it's invalid, update it
			if(!vbo_valid[EL_TEXCOORD]) {
				glBufferDataARB(GL_ARRAY_BUFFER_ARB, ntc * sizeof *tc, tc, GL_DYNAMIC_DRAW_ARB);
				vbo_valid[EL_TEXCOORD] = true;
			}
			glTexCoordPointer(2, GL_FLOAT, 0, 0);
		} else {
			glTexCoordPointer(2, GL_FLOAT, 0, tc);
		}
	}

	// do we have tangents?
	if(tang && caps.glsl) {
		glEnableVertexAttribArrayARB(SDR_ATTR_TANGENT);

		if(caps.vbo) {
			glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo[EL_TANGENT]);

			// if it's invalid, update it
			if(!vbo_valid[EL_TANGENT]) {
				glBufferDataARB(GL_ARRAY_BUFFER_ARB, ntang * sizeof *tang, tang, GL_DYNAMIC_DRAW_ARB);
				vbo_valid[EL_TANGENT] = true;
			}
			glVertexAttribPointerARB(SDR_ATTR_TANGENT, 3, GL_FLOAT, 0, 0, 0);
		} else {
			glVertexAttribPointerARB(SDR_ATTR_TANGENT, 3, GL_FLOAT, 0, 0, tang);
		}
	}

	if(caps.vbo) {
		// restore default binding
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
	}
}


#ifndef DEBUG_DRAWING

void trimesh::draw() const
{
	// if we've got a display list, just use it...
	if(dlist) {
		glCallList(dlist);
		return;
	}

	// if the mesh is NOT dynamic, then start compiling a display list
	if(!dynamic) {
		dlist = glGenLists(1);
		glNewList(dlist, GL_COMPILE);
	}

	setup_vertex_arrays();
	
	if(index) {		// indexed triangles?
		if(caps.vbo) {
			glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbo[EL_INDEX]);

			// update if needed
			if(!vbo_valid[EL_INDEX]) {
				glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, nindex * sizeof *index,
						index, GL_DYNAMIC_DRAW_ARB);
				vbo_valid[EL_INDEX] = true;
			}
			
			glDrawElements(GL_TRIANGLES, nindex, GL_UNSIGNED_INT, 0);
			glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
		} else {
			glDrawElements(GL_TRIANGLES, nindex, GL_UNSIGNED_INT, index);
		}

	} else {
		// raw vertex data, group them by 3 and draw
		glDrawArrays(GL_TRIANGLES, 0, nvert);
	}

	// disable all possible vertex arrays
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	if(caps.glsl) {
		glDisableVertexAttribArrayARB(SDR_ATTR_TANGENT);
	}

	if(!dynamic) {
		glEndList();
		glCallList(dlist);
	}
}

#else

void trimesh::draw() const
{
	glBegin(GL_TRIANGLES);
	if(index) {
		for(int i=0; i<nindex; i++) {
			int idx = index[i];
			glNormal3f(norm[idx].x, norm[idx].y, norm[idx].z);
			glTexCoord2f(tc[idx].x, tc[idx].y);
			glVertex3f(vert[idx].x, vert[idx].y, vert[idx].z);
		}
	} else {
		for(int i=0; i<nvert; i++) {
			glNormal3f(norm[i].x, norm[i].y, norm[i].z);
			glTexCoord2f(tc[i].x, tc[i].y);
			glVertex3f(vert[i].x, vert[i].y, vert[i].z);
		}
	}
	glEnd();
}
#endif

void trimesh::calc_bounds()
{
	centroid = Vector3(0, 0, 0);
	aabb_min = Vector3(FLT_MAX, FLT_MAX, FLT_MAX);
	aabb_max = Vector3(FLT_MIN, FLT_MIN, FLT_MIN);

	for(int i=0; i<nvert; i++) {
		centroid += vert[i];
		aabb_min.x = MIN(aabb_min.x, vert[i].x);
		aabb_max.x = MAX(aabb_max.x, vert[i].x);
		aabb_min.y = MIN(aabb_min.y, vert[i].y);
		aabb_max.y = MAX(aabb_max.y, vert[i].y);
		aabb_min.z = MIN(aabb_min.z, vert[i].z);
		aabb_max.z = MAX(aabb_max.z, vert[i].z);
	}
	centroid /= (float)nvert;

	bounds_valid = true;
}


void trimesh::draw_normals(float sz, const color &col) const
{
	if(!norm) return;

	glBegin(GL_LINES);
	glColor4f(col.x, col.y, col.z, col.w);
	for(int i=0; i<nvert; i++) {
		float x = vert[i].x;
		float y = vert[i].y;
		float z = vert[i].z;

		glVertex3f(x, y, z);
		glVertex3f(x + norm[i].x * sz, y + norm[i].y * sz, z + norm[i].z * sz);
	}
	glEnd();
}


void trimesh::draw_tangents(float sz, const color &col) const
{
	if(!tang) return;

	glBegin(GL_LINES);
	glColor4f(col.x, col.y, col.z, col.w);
	for(int i=0; i<nvert; i++) {
		float x = vert[i].x;
		float y = vert[i].y;
		float z = vert[i].z;

		glVertex3f(x, y, z);
		glVertex3f(x + tang[i].x * sz, y + tang[i].y * sz, z + tang[i].z * sz);
	}
	glEnd();
}

void trimesh::draw_vertices(float sz, const color &col) const
{
	glPushAttrib(GL_POINT_BIT);

	glEnable(GL_POINT_SMOOTH);
	glPointSize(sz);

	glBegin(GL_POINTS);
	glColor4f(col.x, col.y, col.z, col.w);
	for(int i=0; i<nvert; i++) {
		glVertex3f(vert[i].x, vert[i].y, vert[i].z);
	}
	glEnd();
	glPopAttrib();
}
