#include <fstream>
#include <map>
#include <string>
#include <errno.h>
#include <assert.h>
#include "opengl.h"
#include "texture.h"
#include "pixmap.h"
#include "color.h"
#include "errlog.h"
#include "port.h"
#include "henge.h"

using namespace std;
using namespace henge;

namespace henge {
	bool set_render_target_fbo(texture *tex, int mrt_idx);
	bool set_render_target_copy(texture *tex, int mrt_idx);
}

static map<string, texture*> texman;

static stack<texture*> rt;
static unsigned int fbo;


bool henge::init_textures()
{
	if(caps.fbo) {
		glGenFramebuffersEXT(1, &fbo);
	}

	rt.push(0);	// initial render target is the regular framebuffer
	get_viewport(0, 0, 0, 0);	// make sure the internal viewport is updated

	return true;
}

void henge::destroy_textures()
{
	if(caps.fbo) {
		glDeleteFramebuffersEXT(1, &fbo);
	}

	while(rt.size()) {
		rt.pop();
	}

	map<string, texture*>::iterator iter = texman.begin();
	while(iter != texman.end()) {
		delete iter++->second;
	}
	texman.clear();
}

/* texture manager - get_texture
 *
 * Requests a texture, identified by name, from the texture manager.
 *
 * If a texture is not found associated with that name in the database,
 * the name will be treated as a filename and an attempt will be made
 * to load it from the filesystem and add it to the database. Otherwise
 * a pointer to the existing texture will be returned.
 *
 * TODO: At this point textures stay in the texture manager database for ever.
 *	   implement texture sets if the need arises.
 * TODO: Should be fun to encode texture generators in the "name" string and have
 *	   textures be generated on demand and requested through the same interface.
 */
texture *henge::get_texture(const char *name, unsigned int textype)
{
	map<string, texture*>::iterator iter = texman.find(name);
	if(iter != texman.end()) {
		return iter->second;
	}

	texture *tex;

	switch(textype) {
	case TEX_2D:
		tex = new texture_2d;
		break;

	case TEX_CUBE:
		tex = new texture_cube;
		break;

	default:
		// try to guess...
		if(strstr(name, ".cube")) {
			tex = new texture_cube;
		} else {
			tex = new texture_2d;
		}
	}

	if(strstr(name, "texgen:") == name) {
		info("generating texture: %s\n", name + strlen("texgen:"));
		if(!tex->generate(name + strlen("texgen:"))) {
			delete tex;
			return 0;
		}
	} else {
		info("loading texture: %s\n", name);
		if(!tex->load(name)) {
			delete tex;
			return 0;
		}
	}

	add_texture(name, tex);
	return tex;
}

bool henge::add_texture(const char *name, texture *tex)
{
	try {
		texman[name] = tex;
	}
	catch(...) {
		return false;
	}
	return true;
}


bool henge::dump_textures(bool managed_only)
{
#if !defined(WIN32) && !defined(__WIN32__)
#define TEX_DUMP_DIR	"texdump"

	// create the texture dump directory
	if(mkdir(TEX_DUMP_DIR, 0755) == -1) {
		if(errno == EEXIST) {
			struct stat st;
			if(stat(TEX_DUMP_DIR, &st) == -1 || !S_ISDIR(st.st_mode)) {
				return false;
			}
		} else {
			return false;
		}
	}
#else
#define TEX_DUMP_DIR	"."
#endif

	//if(managed_only) {
		map<string, texture*>::iterator iter = texman.begin();
		while(iter != texman.end()) {
			const char *name = iter->first.c_str();
			texture *tex = (iter++)->second;

			char *pname = new char[strlen(TEX_DUMP_DIR) + strlen(name) + 2];
			sprintf(pname, "%s/%s", TEX_DUMP_DIR, name);

			if(tex) tex->save(pname);
			delete [] pname;
		}
		return true;
	/*}

	char *pname = (char*)alloca(strlen(TEX_DUMP_DIR) + 6);
	texture_2d tex;

	for(int i=0; i<2048; i++) {
		if(!glIsTexture(i))
			continue;

		glBindTexture(GL_TEXTURE_2D, i);

		tex.set_tex_id(i);

		int xsz, ysz;
		glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &xsz);
		glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &ysz);
		tex.set_width(xsz);
		tex.set_height(ysz);

		sprintf(pname, "%s/%04d.png", TEX_DUMP_DIR, i);
		tex.save(pname);
	}
	return true;
	*/
}


bool henge::push_render_target()
{
	try {
		if(rt.empty()) {
			rt.push(0);
		} else {
			rt.push(rt.top());
		}
	}
	catch(...) {
		return false;
	}
	return true;
}

bool henge::pop_render_target()
{
	if(rt.empty()) {
		return false;
	}

	rt.pop();
	return set_render_target(rt.top());
}

static bool (*setrt_func[])(texture*, int) = {
	set_render_target_copy,
	set_render_target_fbo
};

bool henge::set_render_target(texture *tex, int mrt_idx)
{
	if(!setrt_func[(int)caps.fbo](tex, mrt_idx)) {
		return false;
	}
	rt.top() = tex;

	return true;
}

#ifndef NDEBUG
static const char *fbstat_str[] = {
	"GL_FRAMEBUFFER_COMPLETE",
	"GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT",
	"GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT",
	"no such fbo error",
	"GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS",
	"GL_FRAMEBUFFER_INCOMPLETE_FORMATS",
	"GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER",
	"GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER",
	"GL_FRAMEBUFFER_UNSUPPORTED"
};
#endif

#define ASSERT_GLERROR	\
	do {	\
		GLenum err;	\
		if((err = glGetError()) != GL_NO_ERROR) {	\
			error("%s: %d - GL ERROR: %s (%d)\n", __FILE__, __LINE__, gluErrorString(err), err);	\
			abort();	\
		}	\
	} while(0)

bool henge::set_render_target_fbo(texture *tex, int mrt_idx)
{
	/*
	static GLenum draw_bufs[] = {
		GL_COLOR_ATTACHMENT0_EXT,
		GL_COLOR_ATTACHMENT1_EXT,
		GL_COLOR_ATTACHMENT2_EXT,
		GL_COLOR_ATTACHMENT3_EXT
	};*/

	if(!caps.mrt && mrt_idx > 0) {
		static bool emmited_error;
		if(!emmited_error) {
			error("set_render_target called with mrt_idx = %d, but MRT is not available!\n", mrt_idx);
			emmited_error = true;
		}
		return false;
	}

	if(!tex) {
		// tex is null, reverting to standard framebuffer
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

		int x, y, xsz, ysz;
		get_viewport(&x, &y, &xsz, &ysz);
		glViewport(x, y, xsz, ysz);
	} else {
		// if this texture doesn't have a depth buffer rendertarget, create it
		if(!tex->rt_depth) {
			glGenRenderbuffersEXT(1, &tex->rt_depth);
			glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, tex->rt_depth);
			glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, tex->xsz, tex->ysz);
		}

		// bind the global fbo and attach the color and depth buffers
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT + mrt_idx,
				GL_TEXTURE_2D, tex->tid, 0);

		if(mrt_idx == 0) {
			// only attach the depth buffer once for mrt rendertargets
			glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
					GL_RENDERBUFFER_EXT, tex->rt_depth);
			// TODO add stencil attachment
		/*} else {
			// if using mrt, we must specify number and nature of buffer attachments
			glDrawBuffersARB(mrt_idx + 1, draw_bufs);*/
		}

#ifndef NDEBUG
		int fbstat = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		if(fbstat != GL_FRAMEBUFFER_COMPLETE_EXT) {
			warning("incomplete framebuffer object: %u (%s)\n", fbo,
					fbstat_str[fbstat - GL_FRAMEBUFFER_COMPLETE_EXT]);
		}
#endif
		glViewport(0, 0, tex->get_useful_width(), tex->get_useful_height());
	}
	return true;
}

/* Set_render_target_copy will try to fake set_render_target by copying the
 * framebuffer to a texture after rendering takes place (when the regular
 * framebuffer is reset). The semantics aren't identical, as the regular render
 * target will be dirty after this. Also although I could make it work with
 * multiple levels of push and pop, I won't bother.
 *
 * Ignores the mrt_idx parameter, if you're trying to do mrt without proper
 * render targets, you're clearly mad.
 */
bool henge::set_render_target_copy(texture *tex, int mrt_idx)
{
	static texture *prev;

	if(prev && prev != tex) {
		prev->copy_frame(0, 0, prev->get_useful_width(), prev->get_useful_height());
	}
	prev = tex;
	return true;
}


void henge::set_texture(const texture *tex, int unit)
{
	if(tex) {
		tex->bind(unit);
	} else {
		if(caps.multitex) {
			glActiveTextureARB(GL_TEXTURE0_ARB + unit);
			glDisable(GL_TEXTURE_1D);
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_TEXTURE_CUBE_MAP_ARB);
			glActiveTextureARB(GL_TEXTURE0_ARB);
		}
	}
}

// -- texture abstract base class

texture::texture()
{
	pixels = 0;
	glGenTextures(1, &tid);
	transp = false;
	set_mipmap_gen(true);
	pixfmt = 4;
	wrap = GL_REPEAT;
	min_filter = GL_LINEAR_MIPMAP_LINEAR;
	mag_filter = GL_LINEAR;
	aniso = 1.0;

	xsz = ysz = 0;
	uxsz = uysz = -1;

	/* a depth buffer fbo attachment will be allocated on demand
	 * when this texture is first bound as a render target
	 */
	rt_depth = 0;
}

texture::~texture()
{
	if(pixels) {
		delete [] pixels;
	}
	glDeleteTextures(1, &tid);

	if(rt_depth) {
		glDeleteRenderbuffersEXT(1, &rt_depth);
	}
}

void texture::set_tex_id(unsigned int id)
{
	tid = id;
}

unsigned int texture::get_tex_id() const
{
	return tid;
}

void texture::set_mipmap_gen(bool genmip)
{
	gen_mipmaps = genmip;
}

bool texture::get_mipmap_gen() const
{
	return gen_mipmaps;
}

bool texture::is_transparent() const
{
	return transp;
}

void texture::set_pixel_format(unsigned int fmt)
{
	pixfmt = fmt;
}

unsigned int texture::get_pixel_format() const
{
	return pixfmt;
}

void texture::set_wrap(unsigned int wrap)
{
	glPushAttrib(GL_TEXTURE_BIT);
	glBindTexture(textype, tid);
	glTexParameteri(textype, GL_TEXTURE_WRAP_S, wrap);
	glTexParameteri(textype, GL_TEXTURE_WRAP_T, wrap);
	glTexParameteri(textype, GL_TEXTURE_WRAP_R, wrap);
	glPopAttrib();

	this->wrap = wrap;
}

unsigned int texture::get_wrap() const
{
	return wrap;
}

void texture::set_filter(unsigned int min, unsigned int mag)
{
	glPushAttrib(GL_TEXTURE_BIT);
	glBindTexture(textype, tid);
	glTexParameteri(textype, GL_TEXTURE_MIN_FILTER, min);
	glTexParameteri(textype, GL_TEXTURE_MAG_FILTER, mag);
	glPopAttrib();

	min_filter = min;
	mag_filter = mag;
}

unsigned int texture::get_min_filter() const
{
	return min_filter;
}


unsigned int texture::get_mag_filter() const
{
	return mag_filter;
}

void texture::set_max_anisotropy(float val)
{
	if(caps.aniso) {
		if(val < 1.0) val = caps.max_aniso;

		glPushAttrib(GL_TEXTURE_BIT);
		glBindTexture(textype, tid);
		glTexParameterf(textype, GL_TEXTURE_MAX_ANISOTROPY_EXT, val);
		glPopAttrib();
		
		aniso = val;
	}
}

float texture::get_max_anisotropy() const
{
	return aniso;
}

void texture::set_width(int w)
{
	xsz = w;
}

void texture::set_height(int h)
{
	ysz = h;
}

int texture::get_width() const
{
	return xsz;
}

int texture::get_height() const
{
	return ysz;
}

void texture::set_useful_area(int w, int h)
{
	uxsz = w;
	uysz = h;
}

int texture::get_useful_width() const
{
	return uxsz == -1 ? xsz : uxsz;
}

int texture::get_useful_height() const
{
	return uysz == -1 ? ysz : uysz;
}

void texture::bind(int tunit) const
{
	static bool dirty_texmat;

	/* if we have useless padding, setup the texture matrix to crop it out
	 * useful for presenting non-pow2 pictures and rtt framebuffer images
	 *
	 * NOTE: if we do this, we're also setting the dirty texmat flag, and we're
	 *       setting the 3rd diagonal element to (float)0xbad. If both these
	 *       conditions hold next time around we'll reset the matrix.
	 */

	if(uxsz != -1 && uysz != -1) {
		glPushAttrib(GL_TRANSFORM_BIT);
		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
		glScalef((float)uxsz / (float)xsz, (float)uysz / (float)ysz, (float)0xbad);
		glPopAttrib();
		//info("DBG tex::bind set uxsz/uysz texmat\n");

		dirty_texmat = true;
	} else {
		if(dirty_texmat) {
			float mat[16];
			glGetFloatv(GL_TEXTURE_MATRIX, mat);

			if(fabs(mat[10] - (float)0xbad) < SMALL_NUMBER) {
				glPushAttrib(GL_TRANSFORM_BIT);
				glMatrixMode(GL_TEXTURE);
				glLoadIdentity();
				glPopAttrib();
				//info("DBG tex::bind unset uxsz/uysz texmat\n");
			}
			dirty_texmat = false;
		}
	}

	if(caps.multitex) {
		glActiveTextureARB(GL_TEXTURE0_ARB + tunit);
	}
	glEnable(textype);
	glBindTexture(textype, tid);

	if(caps.multitex) {
		glActiveTextureARB(GL_TEXTURE0_ARB);
	}
}

// -- 2d texture class

texture_2d::texture_2d()
{
	xsz = ysz = -1;
	textype = GL_TEXTURE_2D;

	set_mipmap_gen(true);

	glPushAttrib(GL_TEXTURE_BIT);
	glBindTexture(textype, tid);
	glTexParameteri(textype, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(textype, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(textype, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(textype, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glPopAttrib();
}

texture_2d::~texture_2d()
{
}

void texture_2d::set_mipmap_gen(bool genmip)
{
	texture::set_mipmap_gen(genmip);

	if(caps.gen_mipmaps) {
		glPushAttrib(GL_TEXTURE_BIT);
		glBindTexture(GL_TEXTURE_2D, tid);
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, genmip ? 1 : 0);
		glPopAttrib();
	}
}

bool texture_2d::load(const char *fname)
{
	pixmap img;
	if(!img.load(fname)) {
		return false;
	}
	set_image(img);
	return true;
}

bool texture_2d::save(const char *fname)
{
	if(memcmp(fname, ".cube", 6) == 0) {
		return false;
	}

	pixmap img;
	if(!img.set_pixels(get_image(), xsz, ysz)) {
		return false;
	}
	return img.save(fname);
}

bool texture_2d::generate(const char *expr)
{
	pixmap img;
	if(!img.generate(expr)) {
		return false;
	}
	set_image(img);
	return true;
}

void texture_2d::set_image(const pixmap &img)
{
	set_image(img.pixels, img.xsz, img.ysz);
}

void texture_2d::set_image(uint32_t *pixels, int x, int y)
{
	glPushAttrib(GL_TEXTURE_BIT);
	glBindTexture(textype, tid);

	if(pixels) {
		/* XXX is this worth it? how often are we updating images?
		 * should I add a flag to stop this from happening on heavy-updated
		 * textures?
		 */
		transp = false;
		for(int i=0; i<x*y; i++) {
			if(unpack_alpha(pixels[i]) < 255) {
				transp = true;
				break;
			}
		}
	}

	if(x != xsz || y != ysz) {
		if(this->pixels) {
			delete [] this->pixels;
			this->pixels = 0;
		}
		xsz = x;
		ysz = y;

		if(pixels && gen_mipmaps && !caps.gen_mipmaps) {
			gluBuild2DMipmaps(GL_TEXTURE_2D, pixfmt, x, y, GLPIXFMT, GL_UNSIGNED_BYTE, pixels);
		} else {
			glTexImage2D(textype, 0, pixfmt, x, y, 0, GLPIXFMT, GL_UNSIGNED_BYTE, pixels);
		}
	} else {
		if(pixels && gen_mipmaps && !caps.gen_mipmaps) {
			gluBuild2DMipmaps(GL_TEXTURE_2D, pixfmt, x, y, GLPIXFMT, GL_UNSIGNED_BYTE, pixels);
		} else {
			// faster version, no need to re-create the texture_2d, just update the pixels
			glTexSubImage2D(textype, 0, 0, 0, x, y, GLPIXFMT, GL_UNSIGNED_BYTE, pixels);
		}
	}
	// NOTE: GLPIXFMT defined in byteorder.h

	glPopAttrib();
}

uint32_t *texture_2d::get_image() const
{
	if(!pixels) {
		try {
			pixels = new uint32_t[xsz * ysz];
		}
		catch(...) {
			return 0;
		}
	} else {
		glPushAttrib(GL_TEXTURE_BIT);
		glBindTexture(textype, tid);
		glGetTexImage(textype, 0, GLPIXFMT, GL_UNSIGNED_BYTE, pixels);
		glPopAttrib();
	}
	return pixels;
}


void texture_2d::copy_frame()
{
	copy_frame(0, 0, xsz, ysz);
}

void texture_2d::copy_frame(int x, int y, int xsz, int ysz)
{
	glPushAttrib(GL_TEXTURE_BIT);
	glBindTexture(GL_TEXTURE_2D, tid);
	glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, x, y, xsz, ysz);
	glPopAttrib();
}

// -- cubemap class

texture_cube::texture_cube()
{
	textype = GL_TEXTURE_CUBE_MAP_ARB;
	transp_cube = 0;

	glPushAttrib(GL_TEXTURE_BIT);
	glBindTexture(textype, tid);
	glTexParameteri(textype, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(textype, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(textype, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	glTexParameteri(textype, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(textype, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glPopAttrib();

	memset(face_tex_inited, 0, sizeof face_tex_inited);
}

texture_cube::~texture_cube()
{
}

bool texture_cube::is_transparent() const
{
	return transp_cube != 0;
}

bool texture_cube::load(const char *fname)
{
	string line;

	ifstream file(fname);
	if(!file.is_open()) {
		return false;
	}

	// check magic
	getline(file, line);
	if(line != "CUBE") {
		file.close();
		return load_cross(fname);
	}

	for(int i=0; i<6; i++) {
		getline(file, line);

		pixmap img;
		if(!img.load(line.c_str())) {
			return false;
		}
		set_image(img.pixels, img.xsz, i + GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB);
	}

	return true;
}

static int get_xoffs(int face, int xsz)
{
	switch(face) {
	case 0:	/* +X */
		return 2 * xsz / 3;
	case 1:	/* -X */
		return 0;
	default:
		break;
	}
	return xsz / 3;	/* everything else (vertical bit) */
}

static int get_yoffs(int face, int ysz)
{
	switch(face) {
	case 2:	/* +Y */
		return 0;
	case 3:	/* -Y */
		return ysz / 2;
	case 5:	/* -Z */
		return 3 * ysz / 4;
	default:
		break;
	}
	return ysz / 4;	/* +X, -X, +Z */
}

/*     +---+      load_cube_cross(filename)
 *     |+Y |      loads a cubemap from an image
 * +---+---+---+  file, assumed to contain the
 * |-X |+Z |+X |  cube face images in a vertical
 * +---+---+---+  cross layout.
 *     |-Y |
 *     +---+ \
 *     |-Z |  } last face rotated 180deg
 *     +---+ /
 */
bool texture_cube::load_cross(const char *fname)
{
	int size;
	uint32_t *cube, *dest, *src;

	pixmap img;
	if(!img.load(fname)) {
		return false;
	}

	if(img.xsz / 3 != img.ysz / 4) {
		return false;
	}
	size = img.ysz / 4;

	cube = new uint32_t[size * size];
	
	for(int i=0; i<6; i++) {
		int x, y;
		int xoffs = get_xoffs(i, img.xsz);
		int yoffs = get_yoffs(i, img.ysz);

		dest = i < 5 ? cube : cube + (size - 1) * size;
		src = img.pixels + (yoffs * img.xsz + xoffs);
		for(y=0; y<size; y++) {
			for(x=0; x<size; x++) {
				int dest_x = i < 5 ? x : size - x - 1;
				dest[dest_x] = *src++;
			}
			if(i < 5) {
				dest += size;
			} else {
				dest -= size;
			}
			src += img.xsz - size;
		}

		set_image(cube, size, GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB + i);
		glassert();
	}

	delete [] cube;
	return true;
}

bool texture_cube::save(const char *fname)
{
	char *face_fname = (char*)alloca(strlen(fname) + 4);

	pixmap img;

	for(int i=0; i<6; i++) {
		uint32_t *pix = get_image(i + GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB);

		char sign = i % 2 ? 'n' : 'p';
		char axis = i / 3 + 'x';
		sprintf(face_fname, "%c%c_%s", sign, axis, fname);

		if(!img.set_pixels(pix, xsz, ysz)) {
			return false;
		}
		if(!img.save(face_fname)) {
			return false;
		}
	}
	return true;
}

bool texture_cube::generate(const char *expr)
{
	return false;	// TODO
}

void texture_cube::set_image(const pixmap &img, unsigned int face)
{
	set_image(img.pixels, img.xsz, face);
}

void texture_cube::set_image(uint32_t *pixels, int size, unsigned int face)
{
	assert(face >= GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB && face <= GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB);

	glPushAttrib(GL_TEXTURE_BIT);
	glBindTexture(textype, tid);

	int fidx = face - GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB;

	transp_cube &= ~(1 << fidx);
	for(int i=0; i<size*size; i++) {
		if(unpack_alpha(pixels[i]) < 255) {
			transp_cube |= 1 << fidx;
			break;
		}
	}

	if(xsz != size) {
		if(this->pixels) {
			delete [] this->pixels;
			this->pixels = 0;
		}
		xsz = size;
		ysz = size;
		memset(face_tex_inited, 0, sizeof face_tex_inited);
	}

	if(!face_tex_inited[fidx]) {
		glTexImage2D(face, 0, pixfmt, size, size, 0, GLPIXFMT, GL_UNSIGNED_BYTE, pixels);
		glassert();
		face_tex_inited[fidx] = true;
	} else {
		// faster version, no need to re-create the texture_2d, just update the pixels
		glTexSubImage2D(face, 0, 0, 0, size, size, GLPIXFMT, GL_UNSIGNED_BYTE, pixels);
		glassert();
	}

	glPopAttrib();
}

uint32_t *texture_cube::get_image(unsigned int face) const
{
	if(!pixels) {
		try {
			pixels = new uint32_t[xsz * ysz];
		}
		catch(...) {
			return 0;
		}
	} else {
		glPushAttrib(GL_TEXTURE_BIT);
		glBindTexture(textype, tid);
		glGetTexImage(face, 0, GLPIXFMT, GL_UNSIGNED_BYTE, pixels);
		glPopAttrib();
	}

	return pixels;
}

int texture_cube::get_size() const
{
	assert(xsz == ysz);
	return xsz;
}

void texture_cube::copy_frame()
{
	copy_frame(0, 0, xsz, ysz, GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB);
}

void texture_cube::copy_frame(int face)
{
	copy_frame(0, 0, xsz, ysz, face);
}

void texture_cube::copy_frame(int x, int y, int xsz, int ysz)
{
	copy_frame(0, 0, xsz, ysz, GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB);
}

void texture_cube::copy_frame(int x, int y, int xsz, int ysz, int face)
{
	glPushAttrib(GL_TEXTURE_BIT);
	glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, tid);
	glCopyTexSubImage2D(face, 0, 0, 0, x, y, xsz, ysz);
	glPopAttrib();
}

void henge::heightmap_to_normalmap(texture_2d *tex, float scale_fact)
{
	uint32_t *src = tex->get_image();
	int xsz = tex->get_width();
	int ysz = tex->get_height();

	uint32_t *norm_pixels = new uint32_t[xsz * ysz];
	uint32_t *dst = norm_pixels;

	for(int j=0; j<ysz; j++) {
		for(int i=0; i<xsz; i++) {
			int x0 = i > 0 ? -1 : 0;
			int y0 = j > 0 ? -xsz : 0;
			int x1 = i < xsz - 1 ? 1 : 0;
			int y1 = j < ysz - 1 ? xsz : 0;

			float dfdx = (float)(src[x0] & 0xff) / 255.0f - (float)(src[x1] & 0xff) / 255.0f;
			float dfdy = (float)(src[y0] & 0xff) / 255.0f - (float)(src[y1] & 0xff) / 255.0f;

			Vector3 n(dfdx * scale_fact, dfdy * scale_fact, 1);
			n.normalize();
			n = (n + 1.0f) * 127.5f;

			*dst++ = pack_color((int)n.x, (int)n.y, (int)n.z, 255);
			src++;
		}
	}

	tex->set_image(norm_pixels, xsz, ysz);
}
