#include "procgraph.h"
#include "timer.h"

using namespace henge;

ProcGraph::ProcGraph()
{
	pos = Vector2(0, 0);
	size = Vector2(1, 1);
	tid = 0;
	tpool = 0;
	samples = 0;
	num_samples = 0;
	memset(&prev_stats, 0, sizeof prev_stats);
	prev_time = 0;
	cur_sample = 0;
}

ProcGraph::~ProcGraph()
{
	delete [] samples;
}

void ProcGraph::set_position(const Vector2 &pos)
{
	this->pos = pos;
}

void ProcGraph::set_size(const Vector2 &size)
{
	this->size = size;
}

bool ProcGraph::init(int tid, const ThreadPool *tpool, int num_samples)
{
	delete [] samples;
	cur_sample = 0;

	samples = new ProcSample[num_samples];
	memset(samples, 0, num_samples * sizeof *samples);
	this->num_samples = num_samples;

	this->tid = tid;
	this->tpool = tpool;

	return true;
}

void ProcGraph::update()
{
	ThreadStats stats = *tpool->get_thread_stats(tid);
	unsigned long cur_time = get_msec();

	unsigned long interval = cur_time - prev_time;
	unsigned long proc, idle;

	if(stats.idle_start) {
		if(stats.idle_start > prev_time) {
			idle = cur_time - stats.idle_start;
		} else {
			idle = interval;
		}
	} else {
		idle = stats.idle_time - prev_stats.idle_time;
	}

	if(stats.proc_start) {
		if(stats.proc_start > prev_time) {
			proc = cur_time - stats.proc_start;
		} else {
			proc = interval;
		}
	} else {
		proc = stats.proc_time - prev_stats.proc_time;
	}

	samples[cur_sample].proc = (float)proc / (float)interval;
	samples[cur_sample].idle = (float)idle / (float)interval;

	cur_sample = (cur_sample + 1) % num_samples;

	prev_stats = stats;
	prev_time = cur_time;
}

#define BORDER	0.01
void ProcGraph::draw() const
{
	extern int xsz, ysz;
	float aspect = (float)xsz / (float)ysz;

	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, aspect, 0, 1, 0, 1);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(pos.x, pos.y, 0);
	glScalef(size.x / 2.0, size.y / 2.0, 1);

	glBegin(GL_QUADS);

	// draw background
	glColor3f(0.25, 0.25, 0.25);
	glVertex2f(-1, -1);
	glVertex2f(1, -1);
	glVertex2f(1, 1);
	glVertex2f(-1, 1);

	int sidx = cur_sample - num_samples;
	if(sidx < 0) sidx += num_samples;

	float dx = (2.0 - 2.0 * BORDER) / (float)num_samples;
	float x = BORDER - 1.0;

	for(int i=0; i<num_samples; i++) {
		float y;

		if(samples[sidx].proc < samples[sidx].idle) {
			y = (1.0 - samples[sidx].idle) * 2.0 - 1.0;
		} else {
			y = samples[sidx].proc * 2.0 - 1.0;
		}
		sidx = (sidx + 1) % num_samples;

		// proc graph
		glColor3f(1.0, 0.8, 0.52);
		glVertex2f(x, -1);
		glVertex2f(x + dx, -1);
		glVertex2f(x + dx, y);
		glVertex2f(x, y);

		// idle graph
		glColor3f(0.62, 0.78, 0.81);
		glVertex2f(x, y);
		glVertex2f(x + dx, y);
		glVertex2f(x + dx, 1.0);
		glVertex2f(x, 1.0);
		
		x += dx;
	}
	glEnd();

	glBegin(GL_LINES);
	glColor3f(1, 1, 1);
	glVertex2f(-1, -1);
	glVertex2f(1, -1);

	glVertex2f(1, -1);
	glVertex2f(1, 1);
	
	glVertex2f(1, 1);
	glVertex2f(-1, 1);
	
	glVertex2f(-1, 1);
	glVertex2f(-1, -1);
	glEnd();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glPopAttrib();
}
