#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include "sock.h"
#include "proto.h"
#include "timer.h"
#include "simobj.h"
#include "geom.h"

#define MSG(x)	#x,
const char *msg_name[] = {
	MESSAGES
	0
};
#undef MSG

const char *get_msg_name(int msg)
{
	if(msg < 0 || msg >= NUM_MESSAGES) {
		return "<invalid message>";
	}
	return msg_name[msg];
}

bool send_msg_header(int s, const MsgHeader *hdr)
{
	unsigned char hdrbuf[2];

	if(hdr->data_sz >= 256) {
		return false;
	}

	hdrbuf[0] = hdr->msg;
	hdrbuf[1] = hdr->data_sz;

	return send(s, (const char*)hdrbuf, 2, 0) == 2;
}

bool recv_msg_header(int s, MsgHeader *hdr, unsigned long timeout)
{
	unsigned char hdrbuf[2];

	if(recv_data(s, hdrbuf, sizeof hdrbuf, timeout) < (int)sizeof hdrbuf) {
		return false;
	}
	hdr->msg = (int)hdrbuf[0];
	hdr->data_sz = (int)hdrbuf[1];
	return true;
}

/* this function blocks until it gets exactly the specified ammount
 * of data through the socket.
 * timeout values:
 *    0    - don't wait at all
 *    < 0  - wait indefinitely
 *    > 0  - wait for the specified ammount of time
 * returns: bytes received
 */
int recv_data(int s, void *buf, int size, unsigned long timeout)
{
	unsigned char *bufptr = (unsigned char*)buf;
	int rdbytes = 0;

	do {
		if(timeout >= 0) {
			struct timeval tv;
			tv.tv_sec = timeout / 1000;
			tv.tv_usec = timeout * 1000 - tv.tv_sec * 1000000;

			fd_set rdset;
			FD_ZERO(&rdset);
			FD_SET(s, &rdset);

			unsigned long start_wait = get_msec();

			int num_ready;
			do {
				num_ready = select(s + 1, &rdset, 0, 0, &tv);
			} while(num_ready == -1 && errno == EINTR);

			if((timeout -= get_msec() - start_wait) < 0) {
				timeout = 0;
			}

			if(num_ready <= 0) {
				return rdbytes;
			}

			if(!FD_ISSET(s, &rdset)) {
				fprintf(stderr, "wtf?\n");
				continue;
			}
		}

		int rd = recv(s, (char*)bufptr, size - rdbytes, 0);
		if(rd <= 0) {
			return rdbytes;
		}
		rdbytes += rd;
		bufptr += rd;
	} while(rdbytes < size);

	return rdbytes;
}


void ping(int s)
{
	MsgHeader hdr = {MSG_PING, 0};
	send_msg_header(s, &hdr);
}

void alive(int s)
{
	MsgHeader hdr = {MSG_ALIVE, 0};
	send_msg_header(s, &hdr);
}

#define BALL_DATA_SZ	6 * sizeof(float) + 1
#define NETW_FLOAT(x)	htonl(*(unsigned int*)&(x))

bool pass_ball(int s, const SimBall *ball)
{
	MsgHeader hdr = {MSG_PASS_BALL, BALL_DATA_SZ};
	if(!send_msg_header(s, &hdr)) {
		return false;
	}
	
	unsigned char buf[BALL_DATA_SZ];
	unsigned int *ptr = (unsigned int*)buf;

	*ptr++ = NETW_FLOAT(ball->pos.x);
	*ptr++ = NETW_FLOAT(ball->pos.y);
	*ptr++ = NETW_FLOAT(ball->pos.z);
	*ptr++ = NETW_FLOAT(ball->vel.x);
	*ptr++ = NETW_FLOAT(ball->vel.y);
	*ptr++ = NETW_FLOAT(ball->vel.z);
	buf[6 * sizeof(float)] = (unsigned char)ball->type;

	return send(s, (const char*)buf, BALL_DATA_SZ, 0) == BALL_DATA_SZ;
}

#define HOST_FLOAT(res, x)	\
	do {									\
		unsigned int val = (unsigned int)x;	\
		val = ntohl(val);					\
		res = *(float*)&val;				\
	} while(0)

void parse_ball_data(const unsigned char *data, SimBall *ball)
{
	unsigned int *ptr = (unsigned int*)data;

	HOST_FLOAT(ball->pos.x, *ptr++);
	HOST_FLOAT(ball->pos.y, *ptr++);
	HOST_FLOAT(ball->pos.z, *ptr++);
	HOST_FLOAT(ball->vel.x, *ptr++);
	HOST_FLOAT(ball->vel.y, *ptr++);
	HOST_FLOAT(ball->vel.z, *ptr++);
	ball->type = (unsigned int)data[6 * sizeof(float)];
}


bool sync(int s, const BallStats *stats)
{
	int num_ints = 3 + stats->num_exits;
	int *stat_data = (int*)alloca(num_ints * sizeof *stat_data);

	stat_data[0] = htonl(stats->num_spawned);
	stat_data[1] = htonl(stats->num_collected);
	stat_data[2] = htonl(stats->num_exits);

	for(int i=0; i<stats->num_exits; i++) {
		stat_data[3 + i] = htonl(stats->exit_counters[i]);
	}

	MsgHeader hdr;
	hdr.msg = MSG_SYNC;
	hdr.data_sz = num_ints * sizeof *stat_data;

	int bytes = 0;
	if(send_msg_header(s, &hdr)) {
		bytes = send(s, (const char*)stat_data, hdr.data_sz, 0);
	}
	return bytes == hdr.data_sz;
}


void parse_sync_data(const unsigned char *data, BallStats *stats)
{
	int *stat_data = (int*)data;

	int prev_exits = stats->num_exits;

	stats->num_spawned = ntohl(*stat_data++);
	stats->num_collected = ntohl(*stat_data++);
	stats->num_exits = ntohl(*stat_data++);

	/* make sure there isn't a bug that'll give us garbage and we'll
	 * just go on to allocate a few gazillion exit counters.
	 */
	assert(stats->num_exits == prev_exits);

	for(int i=0; i<stats->num_exits; i++) {
		stats->exit_counters[i] = ntohl(*stat_data++);
	}
}
