/* POSIX thread condvar (pthread_cond_*) implementation based on:
 * "Strategies for Implementing POSIX Condition Variables on Win32"
 * by Douglas C. Schmidt and Irfan Pyarali.
 */
#include <errno.h>
#include "pthread.h"
#include "impl.h"

int pthread_cond_init(pthread_cond_t *cond, const pthread_condattr_t *attr)
{
	cond->waiters_count = 0;
	cond->was_broadcast = 0;

	if(!(cond->sema = CreateSemaphore(0, 0, 0x7fffffff, 0))) {
		errno = EAGAIN;
		return -1;
	}
	
	InitializeCriticalSection(&cond->waiters_count_lock);

	if(!(cond->waiters_done = CreateEvent(0, 0, 0, 0))) {
		CloseHandle(cond->sema);
		errno = EAGAIN;
		return -1;
	}

	pthread_mutex_init(&cond->mutex, 0);
	return 0;
}

/* TODO: handle EBUSY */
int pthread_cond_destroy(pthread_cond_t *cond)
{
	if(!cond->sema) {
		errno = EINVAL;
		return -1;
	}
	DeleteCriticalSection(&cond->waiters_count_lock);
	CloseHandle(cond->sema);
	CloseHandle(cond->waiters_done);
	pthread_mutex_destroy(&cond->mutex);
	return 0;
}

int pthread_cond_signal(pthread_cond_t *cond)
{
	int have_waiters;

	if(!cond->sema) {
		if(pthread_cond_init(cond, 0) == -1) {
			errno = EINVAL;
			return -1;
		}
	}

	EnterCriticalSection(&cond->waiters_count_lock);
	have_waiters = cond->waiters_count > 0;
	LeaveCriticalSection(&cond->waiters_count_lock);

	if(have_waiters) {
		ReleaseSemaphore(cond->sema, 1, 0);
	}
	return 0;
}

int pthread_cond_broadcast(pthread_cond_t *cond)
{
	int have_waiters = 0;

	if(!cond->sema) {
		if(pthread_cond_init(cond, 0) == -1) {
			errno = EINVAL;
			return -1;
		}
	}

	EnterCriticalSection(&cond->waiters_count_lock);
	if(cond->waiters_count > 0) {
		have_waiters = 1;
		cond->was_broadcast = 1;
	}

	if(have_waiters) {
		ReleaseSemaphore(cond->sema, cond->waiters_count, 0);
		LeaveCriticalSection(&cond->waiters_count_lock);

		WaitForSingleObject(cond->waiters_done, INFINITE);
		cond->was_broadcast = 0;
	} else {
		LeaveCriticalSection(&cond->waiters_count_lock);
	}
	return 0;
}

int pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{
	int last_waiter;

	if(!cond->sema) {
		if(pthread_cond_init(cond, 0) == -1) {
			errno = EINVAL;
			return -1;
		}
	}

	EnterCriticalSection(&cond->waiters_count_lock);
	cond->waiters_count++;
	LeaveCriticalSection(&cond->waiters_count_lock);

	/* This call atomically releases the mutex and waits on the
	 * semaphore until pthread_cond_signal or pthread_cond_broadcast
	 * are called by another thread.
	 */
	SignalObjectAndWait(mutex->mut, cond->sema, INFINITE, 0);

	EnterCriticalSection(&cond->waiters_count_lock);
	cond->waiters_count--;

	/* Check to see if we're the last waiter after pthread_cond_broadcast */
	last_waiter = cond->was_broadcast && !cond->waiters_count;
	LeaveCriticalSection(&cond->waiters_count_lock);

	/* If we're the last waiter thread during this particular broadcast
	 * then let all the other threads proceed.
	 */
	if(last_waiter) {
		/* This call atomically signals the waiters_done event and waits until
		 * it can acquire the external_mutex. This is required to ensure fairness.
		 */
		SignalObjectAndWait(cond->waiters_done, mutex->mut, INFINITE, FALSE);
	} else {
		/* Always regain the external mutex */
		WaitForSingleObject(mutex->mut, INFINITE);
	}
	return 0;
}

int pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex, const struct timespec *abstime)
{
	int last_waiter;
	unsigned int wait_msec = abstime->tv_sec * 1000 + (unsigned int)(abstime->tv_nsec / 1000000L);

	if(!cond->sema) {
		if(pthread_cond_init(cond, 0) == -1) {
			errno = EINVAL;
			return -1;
		}
	}

	EnterCriticalSection(&cond->waiters_count_lock);
	cond->waiters_count++;
	LeaveCriticalSection(&cond->waiters_count_lock);

	/* This call atomically releases the mutex and waits on the
	 * semaphore until pthread_cond_signal or pthread_cond_broadcast
	 * are called by another thread.
	 */
	SignalObjectAndWait(mutex->mut, cond->sema, wait_msec, 0);

	EnterCriticalSection(&cond->waiters_count_lock);
	cond->waiters_count--;

	/* Check to see if we're the last waiter after pthread_cond_broadcast */
	last_waiter = cond->was_broadcast && !cond->waiters_count;
	LeaveCriticalSection(&cond->waiters_count_lock);

	/* If we're the last waiter thread during this particular broadcast
	 * then let all the other threads proceed.
	 */
	if(last_waiter) {
		/* This call atomically signals the waiters_done event and waits until
		 * it can acquire the external_mutex. This is required to ensure fairness.
		 */
		SignalObjectAndWait(cond->waiters_done, mutex->mut, INFINITE, FALSE);
	} else {
		/* Always regain the external mutex */
		WaitForSingleObject(mutex->mut, INFINITE);
	}
	return 0;
}
