#if !defined(PTHREAD_H_) && (defined(WIN32) || defined(__WIN32__))
#define PTHREAD_H_

#include <time.h>
#include <windows.h>

/* detach state */
enum {
	PTHREAD_CREATE_JOINABLE,
	PTHREAD_CREATE_DETACHED
};

/* cancellation state */
enum {
	PTHREAD_CANCEL_ENABLE,
	PTHREAD_CANCEL_DISABLE
};

/* cancellation type */
enum {
	PTHREAD_CANCEL_DEFERRED,
	PTHREAD_CANCEL_ASYNCHRONOUS
};

struct sched_param {
	int __sched_priority;
};

struct timespec {
	time_t tv_sec;
	long tv_nsec;
};

#define PTHREAD_MUTEX_INITIALIZER	{0}
#define PTHREAD_COND_INITIALIZER	{0, {0, 0, 0, 0, 0, 0}, 0, 0, 0, {0}}

typedef int pthread_t;

typedef struct {
	int detached;
	size_t guard_size;			/* ignored */
	struct sched_param sched;	/* ignored */
	int sched_policy;			/* ignored */
	int inherit_sched;			/* ignored */
	int scope;					/* ignored */
	size_t stack_size;
} pthread_attr_t;

typedef struct {
	HANDLE mut;
} pthread_mutex_t;

typedef struct { int foo; } pthread_mutexattr_t;

typedef struct {
	int waiters_count;	/* Number of waiting threads */
	CRITICAL_SECTION waiters_count_lock;
	
	/* Semaphore used to queue up threads waiting for the condition
	 * to become signaled
	 */
	HANDLE sema;

	/* An auto-reset event used by the broadcast/signal thread to wait
	 * for all the waiting thread(s) to wake up and be released from
	 * the semaphore
	 */
	HANDLE waiters_done;
	
	/* Keeps track of whether we were broadcasting or signaling.
	 * This allows us to optimize the code if we're just signaling.
	 */
	size_t was_broadcast;

	pthread_mutex_t mutex;	/* to be held inside cond functions */
} pthread_cond_t;

typedef struct { int foo; } pthread_condattr_t;

typedef DWORD pthread_key_t;
typedef int pthread_once_t;

#ifdef __cplusplus
extern "C" {
#endif

int pthread_create(pthread_t *tid, const pthread_attr_t *attr,
		void *(*tfunc)(void*), void *arg);

void pthread_exit(void *val_ptr);

int pthread_join(pthread_t tid, void **val_ptr);

int pthread_detach(pthread_t tid);

pthread_t pthread_self(void);

int pthread_equal(pthread_t t1, pthread_t t2);

/* pthread attributes */
int pthread_attr_init(pthread_attr_t *attr);
int pthread_attr_destroy(pthread_attr_t *attr);

int pthread_attr_getdetachstate(const pthread_attr_t *attr, int *dstate);
int pthread_attr_setdetachstate(pthread_attr_t *attr, int dstate);

int pthread_attr_getguardsize(const pthread_attr_t *attr, size_t *gsize);
int pthread_attr_setguardsize(pthread_attr_t *attr, size_t gsize);

int pthread_attr_getschedparam(const pthread_attr_t *attr, struct sched_param *param);
int pthread_attr_setschedparam(pthread_attr_t *attr, const struct sched_param *param);

int pthread_attr_getschedpolicy(const pthread_attr_t *attr, int *policy);
int pthread_attr_setschedpolicy(pthread_attr_t *attr, int policy);

int pthread_attr_getinheritsched(const pthread_attr_t *attr, int *inherit);
int pthread_attr_setinheritsched(pthread_attr_t *attr, int inherit);

int pthread_attr_getscope(const pthread_attr_t *attr, int *scope);
int pthread_attr_setscope(pthread_attr_t *attr, int scope);

int pthread_attr_getstacksize(const pthread_attr_t *attr, size_t *ssize);
int pthread_attr_setstacksize(pthread_attr_t *attr, size_t ssize);

/* scheduler parameters */
/*
int pthread_setschedparam(pthread_t tid, int policy, const struct sched_param *param);
int pthread_getschedparam(pthread_t tid, int *policy, struct sched_param *param);
*/

/* set thread priority */
int pthread_setschedprio(pthread_t tid, int prio);

/* cancellation */
int pthread_setcancelstate(int state, int *oldstate);
int pthread_setcanceltype(int type, int *oldtype);
int pthread_cancel(pthread_t tid);
void pthread_testcancel(void);

/* TODO missing pthread_cleanup_push/pop */

/* mutexen */
int pthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *mattr);
int pthread_mutex_destroy(pthread_mutex_t *mutex);

int pthread_mutex_trylock(pthread_mutex_t *mutex);
int pthread_mutex_lock(pthread_mutex_t *mutex);
int pthread_mutex_unlock(pthread_mutex_t *mutex);

/* mutex attributes */
int pthread_mutexattr_init(pthread_mutexattr_t *attr);
int pthread_mutexattr_destroy(pthread_mutexattr_t *attr);

int pthread_mutexattr_getpshared(const pthread_mutexattr_t *attr, int *pshared);
int pthread_mutexattr_setpshared(pthread_mutexattr_t *attr, int pshared);

/* conditions */
int pthread_cond_init(pthread_cond_t *cond, const pthread_condattr_t *attr);
int pthread_cond_destroy(pthread_cond_t *cond);

int pthread_cond_signal(pthread_cond_t *cond);
int pthread_cond_broadcast(pthread_cond_t *cond);

int pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex);
int pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex,
		const struct timespec *abstime);

/* condition attributes */
int pthread_condattr_init(pthread_condattr_t *attr);
int pthread_condattr_destroy(pthread_condattr_t *attr);

int pthread_condattr_getpshared(const pthread_condattr_t *attr, int *pshared);
int pthread_condattr_setpshared(pthread_condattr_t *attr, int pshared);

/* thread-specific data */
int pthread_key_create(pthread_key_t *key, void (*destr_func)(void*));
int pthread_key_delete(pthread_key_t key);

/*void *pthread_getspecific(pthread_key_t key);
int pthread_setspecific(pthread_key_t key, const void *ptr);*/

#define pthread_getspecific(key) 		TlsGetValue(key)
#define pthread_setspecific(key, ptr)	(TlsSetValue(key, ptr) ? 0 : -1)

/* XXX not going to implement pthread_atfork (meaningless on win32) */
/* int pthread_atfork(void (*prepare)(void), void (*parent)(void), void (*child)(void)); */

#ifdef __cplusplus
}
#endif

#endif	/* !PTHREAD_H_ && (WIN32 || __WIN32__) */
