#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>
#include "sock.h"
#include "server.h"
#include "proto.h"
#include "geom.h"
#include "galton.h"
#include "timer.h"

#if defined(WIN32) || defined(__WIN32__)
#define close(s)	closesocket(s)
#endif

static void *accept_thread(void *arg);
static void handle_client(int s);
static void bounds_callback(const SimBall *ball, void *cls);

static pthread_t acc_tid;
static pthread_mutex_t sock_mutex = PTHREAD_MUTEX_INITIALIZER;
static int sock, lis_sock = -1;
static volatile bool srv_done;

bool init_server(int port)
{
	int s;

	printf("starting up server (port: %d) ...", port);
	fflush(stdout);

	if((s = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("failed to create socket");
		return false;
	}

	struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = INADDR_ANY;

	if(bind(s, (struct sockaddr*)&addr, sizeof addr) == -1) {
		fprintf(stderr, "cannot bind socket to port: %d: %s\n", port, strerror(errno));
		close(s);
		return false;
	}

	if(listen(s, 5) == -1) {
		perror("listen failed");
		close(s);
		return false;
	}

	// spawn the "accept" thread
	lis_sock = s;
	if(pthread_create(&acc_tid, 0, accept_thread, 0) != 0) {
		perror("failed to spawn \"accept\" thread");
		close(s);
		lis_sock = -1;
		return false;
	}

	printf("done\n");
	return true;
}

void shutdown_server()
{
	if(lis_sock != -1) {
		printf("waiting for server to shut down ...");
		fflush(stdout);

		srv_done = true;
		pthread_join(acc_tid, 0);

		close(lis_sock);
		lis_sock = -1;

		printf("done\n");
	}
}

bool send_spawn_message()
{
	if(sock == -1) {
		return false;
	}

	MsgHeader hdr = {MSG_SPAWN, 0};

	pthread_mutex_lock(&sock_mutex);
	bool res = send_msg_header(sock, &hdr);
	pthread_mutex_unlock(&sock_mutex);

	return res;
}

static void *accept_thread(void *arg)
{
	while(!srv_done) {
		struct timeval tv;
		tv.tv_sec = 2;
		tv.tv_usec = 0;

		fd_set rdset;
		FD_ZERO(&rdset);
		FD_SET(lis_sock, &rdset);

		int num_rd;
		do {
			num_rd = select(lis_sock + 1, &rdset, 0, 0, &tv);
		} while(num_rd == -1 && errno == EINTR);

		if(num_rd == -1) {
			perror("select failed");
			continue;
		}

		if(num_rd > 0 && FD_ISSET(lis_sock, &rdset)) {
			// we got a client
			int s;
			if((s = accept(lis_sock, 0, 0)) == -1) {
				perror("accept failed");
				continue;
			}

			/* modify the simulation limits, to get rid of the balls when they go
			 * to the bottom half, and get a callback called so that we may pass them
			 * to the client.
			 */
			double lim_upper, lim_lower;
			double half_point = box_param.rows * box_param.row_dist / 2.0;
			sim->get_limits(&lim_lower, &lim_upper);
			sim->set_limits(half_point, lim_upper);
			sim->set_bounds_callback(bounds_callback, &s);

			sock = s;
			// we only talk to one client at a time, no need for concurrency
			handle_client(s);

			// cleanup, we no longer have a client
			sock = -1;
			sim->set_limits(lim_lower, lim_upper);
			close(s);
		}
	}
	return 0;
}

static void handle_client(int s)
{
	unsigned long ping_sent = 0;
	MsgHeader hdr;
	const char *greet_str = "GALTON/1.0";

	hdr.msg = MSG_HELLO;
	hdr.data_sz = strlen(greet_str);
	if(!send_msg_header(s, &hdr) || send(s, greet_str, hdr.data_sz, 0) < hdr.data_sz) {
		close(s);
		return;
	}

	sync(s, &ball_stats);

	while(!srv_done) {
		unsigned char data_buf[256];

		if(ping_sent) {
			if(get_msec() - ping_sent > PING_TIMEOUT) {
				printf("ping timeout...\n");
				break;
			}
		} else {
			pthread_mutex_lock(&sock_mutex);
			printf("ping...\n");
			ping(sock);
			pthread_mutex_unlock(&sock_mutex);

			ping_sent = get_msec();
		}

		// get message
		if(!recv_msg_header(sock, &hdr, TIMEOUT)) {
			continue;
		}
		// get the message data into the data buffer
		if(recv_data(sock, data_buf, hdr.data_sz, TIMEOUT) < hdr.data_sz) {
			printf("huh?\n");
			continue;
		}

		printf("got message: %s\n", get_msg_name(hdr.msg));

		switch(hdr.msg) {
		case MSG_ALIVE:
			ping_sent = 0;
			break;

		case MSG_COLLECT:
			{
				pthread_mutex_lock(&ball_stats.mutex);
				ball_stats.num_collected++;
			
				int exit_slot = (int)data_buf[0];
				if(exit_slot >= ball_stats.num_exits) {
					fprintf(stderr, "invalid exit slot in COLLECT message: %d\n", exit_slot);
				} else {
					ball_stats.exit_counters[exit_slot]++;
				}
				pthread_mutex_unlock(&ball_stats.mutex);
			}
			break;

		case MSG_PASS_BALL:
			{
				SimBall *ball = new SimBall;
				parse_ball_data(data_buf, ball);

				ball->rad = get_ball_radius(ball->type);
				ball->mass = get_ball_mass(ball->type);
				ball->rest = get_ball_restitution(ball->type);
				sim->add_ball(ball);
			}
			break;

		case MSG_PING:
			pthread_mutex_lock(&sock_mutex);
			alive(s);
			pthread_mutex_unlock(&sock_mutex);
			break;

		case MSG_BYE:
			return;

		default:
			fprintf(stderr, "unexpected message: %s\n", get_msg_name(hdr.msg));
		}
	}
}

/* This function is called by Simulation whenever a ball escapes the specified
 * upper or lower limits. Our task is to see if that was the lower limit (half-way
 * point of the box) and if so, pass the ball to the client.
 */
static void bounds_callback(const SimBall *ball, void *cls)
{
	double upper, lower;
	sim->get_limits(&lower, &upper);

	if(sock != -1 && ball->pos.y < lower) {
		printf("passing ball: [%.3f, %.3f, %.3f]\n", ball->pos.x, ball->pos.y, ball->pos.z);

		pthread_mutex_lock(&sock_mutex);
		pass_ball(sock, ball);
		pthread_mutex_unlock(&sock_mutex);
	}
}
