#include <map>
#include <float.h>
#include <assert.h>
#if defined(WIN32) || defined(__WIN32__)
#include <malloc.h>
#else
#include <alloca.h>
#endif
#include <pthread.h>
#include "sim.h"
#include "galton.h"
#include "timer.h"

#define MAX_COL_STEPS	1000

using namespace std;

struct TaskData {
	double dt;
	Simulation *sim;
	list<SimBall*> ball_list;
};

static void *sim_loop(void *arg);
static void task_integrate(void *cls);
static void task_collide(void *cls);

static bool aabb_vert_test(const SimBall *ball, const AABox &aabox, double mov_dist_sq);
static void calc_response(SimBall *ball_a, SimBall *ball_b, const Contact *cont);
static bool collision_handled(SimBall *ball_a, SimBall *ball_b);

static map<SimBall*, SimBall*> colset;
static pthread_mutex_t colset_mutex = PTHREAD_MUTEX_INITIALIZER;

extern SimStaticObj *ground_sim_obj;

Simulation::Simulation()
{
	time_scaling = 1.0;
	num_balls = 0;

	timeslice = prev_upd = 0.0025;
	grav = Vector3(0.0, -9.8, 0.0);

	upper_limit = DBL_MAX;
	lower_limit = -40.0;
	bounds_callback = 0;
	bounds_callback_cls = 0;

	acc_col = acc_cont = false;

	drag_coef = 0.0;
	
	ball_vis.resize(50);
	for(int i=0; i<50; i++) {
		ball_vis[i].type = BTYPE_INVALID;
	}
	num_vis_balls = 0;

	pthread_mutex_init(&ballvis_mutex, 0);

	int num_threads = 0;	// default: autodetect number of CPUs
	cfg.getopt_int("threads", &num_threads);

	running = false;

	// start the thread pool
	tpool.start(num_threads);
}

Simulation::~Simulation()
{
	stop();
}

bool Simulation::start()
{
	if(running) {
		return false;
	}

	start_pause = 0;
	stoppage = get_msec();

	running = true;
	if(pthread_create(&simloop_thread, 0, sim_loop, this) != 0) {
		return false;
	}

	return true;
}

void Simulation::stop()
{
	if(running) {
		printf("stopping the simulation loop ...");
		fflush(stdout);

		running = false;
		pthread_join(simloop_thread, 0);

		printf("done\n");
	}
}

void Simulation::toggle_pause()
{
	if(start_pause > 0) {	// unpause
		stoppage += get_msec() - start_pause;
		start_pause = 0;
	} else {
		start_pause = get_msec();
	}
}

bool Simulation::is_paused() const
{
	return start_pause > 0;
}

double Simulation::get_time() const
{
	unsigned long msec;
	if(start_pause) {
		msec = start_pause - stoppage;
	} else {
		msec = get_msec() - stoppage;
	}
	return (double)msec / 1000.0;
}

void Simulation::set_time_scaling(double s)
{
	time_scaling = s;
}

void Simulation::set_collision_gathering(bool enable)
{
	acc_col = enable;
}

void Simulation::set_contact_gathering(bool enable)
{
	acc_cont = enable;
}

void Simulation::set_limits(double lower, double upper)
{
	upper_limit = upper;
	lower_limit = lower;

	// go through all the static objects and enable/disable them
	for(size_t i=0; i<sobjects.size(); i++) {
		AABox aabox = sobjects[i]->get_bounds();

		if(aabox.min.y > upper || aabox.max.y < lower) {
			sobjects[i]->deactivate();
		} else {
			sobjects[i]->activate();
		}
	}
}

void Simulation::get_limits(double *lower, double *upper) const
{
	*lower = lower_limit;
	*upper = upper_limit;
}

void Simulation::set_bounds_callback(void (*func)(const SimBall*, void*), void *cls)
{
	bounds_callback = func;
	bounds_callback_cls = cls;
}

void Simulation::set_timeslice(double timeslice)
{
	this->timeslice = timeslice;
}

double Simulation::get_timeslice() const
{
	return timeslice;
}

void Simulation::set_gravity(const Vector3 &grav)
{
	this->grav = grav;
}

const Vector3 &Simulation::get_gravity() const
{
	return grav;
}

void Simulation::set_drag(double drag)
{
	drag_coef = drag;
}

double Simulation::get_drag() const
{
	return drag_coef;
}

void Simulation::add_ball(SimBall *ball)
{
	ball_list.push_back(ball);
	num_balls++;
}

void Simulation::add_static(SimStaticObj *sobj)
{
	sobjects.push_back(sobj);
}

int Simulation::get_ball_count() const
{
	return num_balls;
}

double Simulation::run(double time)
{
	double interval = time - prev_upd;

	if(interval < timeslice) {
		return prev_upd;
	}

	time = prev_upd + timeslice;
	while(interval > timeslice) {
		step(timeslice * time_scaling);
		interval -= timeslice;
		time += timeslice;
	}
	prev_upd = time;

	return time;
}


static void *sim_loop(void *arg)
{
	Simulation *sim = (Simulation*)arg;
	double dt = sim->get_timeslice();

	double prev_upd = -dt;
	while(sim->is_running()) {
		double sec = sim->get_time();
		long time_left = (long)((dt - (sec - prev_upd)) * 1000.0);

		if(time_left > 0) {
			msleep(time_left);
		}

		prev_upd = sim->run(sec);
		sim->remove_dead_balls();
	}

	return 0;
}


static void task_integrate(void *cls)
{
	Simulation *sim = ((TaskData*)cls)->sim;
	list<SimBall*> *ball_list = &((TaskData*)cls)->ball_list;
	double dt = ((TaskData*)cls)->dt;

	list<SimBall*>::iterator iter = ball_list->begin();
	while(iter != ball_list->end()) {
		SimBall *ball = *iter++;

		Vector3 grav = sim->get_gravity();
		double drag_coef = sim->get_drag();

		// find all resting contacts
		list<Contact> cont_list;
		sim->find_contacts(ball, &cont_list);

		// hack for ground contacts :)
		list<Contact>::iterator cont = cont_list.begin();
		while(cont != cont_list.end()) {
			if(cont->sobj == ground_sim_obj) {
				Vector3 reaction = cont->normal;
				if(cont->sobj == ground_sim_obj) {
					reaction += -grav * ball->mass;
				}
				ball->forces += reaction;
			}
			cont++;
		}

		// sum forces, integrate velocity and calculate temporary full-dt position
		Vector3 forces = grav * ball->mass + ball->forces - ball->vel * drag_coef;

		Vector3 accel = forces / ball->mass;
		ball->prev_vel = ball->vel;
		ball->vel += accel * dt;

		ball->end = ball->pos + ball->vel * dt;
	}
}

static void task_collide(void *cls)
{
	Simulation *sim = ((TaskData*)cls)->sim;
	list<SimBall*> *ball_list = &((TaskData*)cls)->ball_list;
	double dt = ((TaskData*)cls)->dt;

	list<SimBall*>::iterator iter = ball_list->begin();
	while(iter != ball_list->end()) {
		SimBall *ball = *iter++;

		Vector3 grav = sim->get_gravity();
		double drag_coef = sim->get_drag();
			
		list<Contact> cont_list;
		int num_col = sim->find_collisions(ball, &cont_list);

		Contact c;
		list<Contact>::iterator cont = cont_list.begin();
		while(cont != cont_list.end()) {
			c = *cont++;
			calc_response(ball, c.ball2, &c);
		}

		if(num_col) {
			double last_cont_dt = c.t * dt;
			ball->pos += ball->prev_vel * last_cont_dt;
		
			double dt_left = (1.0 - c.t) * dt;
			
			Vector3 forces = grav * ball->mass + ball->forces - ball->vel * drag_coef;
			Vector3 accel = forces / ball->mass;
			ball->vel = ball->vel + accel * dt_left;
			ball->pos = ball->pos + ball->vel * dt_left;
		} else {
			ball->pos = ball->end;
		}

		ball->forces = Vector3(0.0, 0.0, 0.0);
	}
}

#define BATCH_SIZE	10
void Simulation::step(double dt)
{
	//Task *tasks = (Task*)alloca(num_balls * sizeof *tasks);
	//TaskData *task_data = new TaskData[num_balls];

	int num_threads = tpool.get_num_threads();
	int batch_size = num_balls < BATCH_SIZE * num_threads ? num_balls / num_threads : BATCH_SIZE;
	if(!batch_size) batch_size = 1;

	int num_tasks = num_balls / batch_size;
	if(num_tasks * batch_size < num_balls) {
		num_tasks++;
	}

	if(!num_tasks) return;

	Task *tasks = (Task*)alloca(num_tasks * sizeof *tasks);
	TaskData *task_data = new TaskData[num_tasks];

	for(int i=0; i<num_tasks; i++) {
		task_data[i].dt = dt;
		task_data[i].sim = this;
		tasks[i].proc = task_integrate;
		tasks[i].done = 0;
		tasks[i].closure = task_data + i;
	}

	int i = 0;
	list<SimBall*>::iterator iter = ball_list.begin();
	while(iter != ball_list.end()) {
		SimBall *ball = *iter++;

		task_data[i].ball_list.push_back(ball);
		if(i >= batch_size) {
			i++;
		}
	}

	tpool.add_work(tasks, num_tasks);
	tpool.wait_work();

	for(int i=0; i<num_tasks; i++) {
		tasks[i].proc = task_collide;
	}
	tpool.add_work(tasks, num_tasks);
	tpool.wait_work();

	colset.clear();
	update_ball_vis();

	delete [] task_data;
}


int Simulation::find_all_collisions(list<Contact> *cont_list)
{
	int num_col = 0;
	cont_list->clear();

	list<SimBall*>::iterator iter = ball_list.begin();
	while(iter != ball_list.end()) {
		SimBall *ball = *iter++;
		num_col += find_collisions(ball, cont_list, true);
	}

	cont_list->sort();

	return num_col;
}

static bool aabb_vert_test(const SimBall *ball, const AABox &aabox, double mov_dist_sq)
{
	double dy1 = ball->pos.y - aabox.max.y - ball->rad;
	double dy2 = aabox.min.y - ball->pos.y - ball->rad;

	if(dy1 > 0.0 && SQ(dy1) > mov_dist_sq) {
		return false;
	}
	if(dy2 > 0.0 && SQ(dy2) > mov_dist_sq) {
		return false;
	}
	return true;
}

int Simulation::find_collisions(SimBall *ball, std::list<Contact> *cont_list, bool append)
{
	int num_col = 0;
	if(!append) {
		cont_list->clear();
	}

	// if it doesn't move it won't collide
	double mov_dist_sq = (ball->pos - ball->end).length_sq();
	if(mov_dist_sq < ERROR_MARGIN) {
		return 0;
	}

	// fist try all active static objects within a vertical window
	for(size_t i=0; i<sobjects.size(); i++) {
		if(!sobjects[i]->is_active()) {
			continue;
		}

		if(!aabb_vert_test(ball, sobjects[i]->get_bounds(), mov_dist_sq)) {
			continue;
		}

		Contact c;
		if(sobjects[i]->intersect(ball, &c)) {
			cont_list->push_back(c);
			num_col++;

			if(acc_col) {
				col_acc_list.push_back(c);
			}
		}
	}

	// then try all other balls
	list<SimBall*>::iterator iter2 = ball_list.begin();
	while(iter2 != ball_list.end()) {
		SimBall *ball2 = *iter2++;

		if(ball2 == ball) {
			continue;
		}

		Contact c;
		if(ball_intersect(ball, ball2, &c)) {
			cont_list->push_back(c);
			num_col++;

			if(acc_col) {
				col_acc_list.push_back(c);
			}
		}
	}

	// XXX ? XXX
	num_col += find_contacts(ball, cont_list, true);

	if(!append) {
		cont_list->sort();
	}
	return num_col;
}

int Simulation::find_contacts(SimBall *ball, list<Contact> *cont_list, bool append)
{
	int num_contacts = 0;

	if(!append) {
		cont_list->clear();
	}

	// first check static objects
	for(size_t i=0; i<sobjects.size(); i++) {
		if(!sobjects[i]->is_active()) {
			continue;
		}

		if(!aabb_vert_test(ball, sobjects[i]->get_bounds(), SQ(ball->rad))) {
			continue;
		}

		Contact cont;
		if(sobjects[i]->contact(ball, &cont)) {
			cont_list->push_back(cont);
			num_contacts++;

			if(acc_cont) {
				cont_acc_list.push_back(cont);
			}
		}
	}

	// then all the other balls
	list<SimBall*>::iterator iter = ball_list.begin();
	while(iter != ball_list.end()) {
		SimBall *ball2 = *iter++;

		if(ball2 == ball) {
			continue;
		}

		Contact cont;
		if(ball_contact(ball, ball2, &cont)) {
			cont_list->push_back(cont);
			num_contacts++;

			if(acc_cont) {
				cont_acc_list.push_back(cont);
			}
		}
	}

	return num_contacts;
}


static void calc_response(SimBall *ball_a, SimBall *ball_b, const Contact *cont)
{
	if(!ball_b) {
		ball_a->vel -= (1.0 + ball_a->rest) * dot_product(ball_a->vel, cont->normal) * cont->normal;

	} else {
		float e = ball_a->rest * ball_b->rest;
		float m1 = ball_a->mass;
		float m2 = ball_b->mass;

		if(collision_handled(ball_a, ball_b)) {
			return;
		}

		//cout << "resp(" << ball_a << ", " << ball_b << ")\n";

		//Vector3 v1 = ball_a->vel.reflection(cont->normal);
		//Vector3 v2 = ball_b->vel.reflection(cont->normal);
		
		//Vector3 v1 = -((1 + e) * dot_product(ball_a->vel, cont->normal) * cont->normal - ball_a->vel);
		//Vector3 v2 = -((1 + e) * dot_product(ball_b->vel, cont->normal) * cont->normal - ball_b->vel);


		Vector3 v1norm = dot_product(ball_a->vel, cont->normal) * cont->normal;
		Vector3 v2norm = dot_product(ball_b->vel, cont->normal) * cont->normal;
		
		//Vector3 v1 = ((m1 - e * m2) * v1norm + (m2 + e * m2) * v2norm) / (m1 + m2);
		//Vector3 v2 = ((m1 + e * m1) * v1norm + (m2 - e * m1) * v2norm) / (m1 + m2);

		Vector3 v1 = ((v1norm * m1) + (v2norm * m2) - (v1norm - v2norm) * m2) / (m1 + m2);
		Vector3 v2 = ((v1norm * m1) + (v2norm * m2) - (v2norm - v1norm) * m1) / (m1 + m2);

		Vector3 v1tang = ball_a->vel - v1norm;
		Vector3 v2tang = ball_b->vel - v2norm;

		ball_a->vel = v1 * e + v1tang * 0.95;
		ball_b->vel = v2 * e + v2tang * 0.95;

		// add to handled collision set
		pthread_mutex_lock(&colset_mutex);
		colset[ball_a] = ball_b;
		colset[ball_b] = ball_a;
		pthread_mutex_unlock(&colset_mutex);
	}
}

static bool collision_handled(SimBall *b1, SimBall *b2)
{
	bool res = false;

	pthread_mutex_lock(&colset_mutex);
	map<SimBall*, SimBall*>::const_iterator iter = colset.find(b1);
	if(iter != colset.end() && iter->second == b2) {
		res = true;
	}
	pthread_mutex_unlock(&colset_mutex);
	return res;
}

/*static void calc_response(SimBall *ball_a, SimBall *ball_b, const Contact *cont)
{
	double mass_rcp_sum, coef_rest = ball_a->rest;
	Vector3 rel_vel = ball_a->vel;

	if(ball_b) {
		coef_rest *= ball_b->rest;
		rel_vel -= ball_b->vel;

		mass_rcp_sum = 1.0 / ball_a->mass + 1.0 / ball_b->mass;
	} else {
		mass_rcp_sum = 1.0 / ball_a->mass;
	}

	Vector3 n = cont->normal;
	double impulse = -((1.0 + coef_rest) * dot_product(rel_vel, n)) / mass_rcp_sum;

	// apply impulse to ball A
	//cout << ball_a->vel << " --- ";
	ball_a->vel += (impulse / ball_a->mass) * n;
	//cout << ball_a->vel << endl;

	// if there is a ball B involved, calculate impulse and apply it also
	if(ball_b) {
		rel_vel = ball_b->vel - ball_a->vel;

		cout << "impulses: " << impulse;
		impulse = -((1.0 + coef_rest) * dot_product(rel_vel, n)) / mass_rcp_sum;
		cout << ", " << impulse;

		// apply impulse
		ball_b->vel += 2.0 * (impulse / ball_b->mass) * n;

		cout << " velocities: " << ball_a->vel.y << ", " << ball_b->vel.y << endl;
	}
}*/

void Simulation::update_ball_vis()
{
	pthread_mutex_lock(&ballvis_mutex);
	if((int)ball_vis.size() < num_balls) {
		ball_vis.resize(num_balls * 2);
	}

	int i = 0;
	list<SimBall*>::iterator iter = ball_list.begin();
	while(iter != ball_list.end()) {
		SimBall *ball = *iter++;

		ball_vis[i].type = ball->type;
		ball_vis[i++].pos = ball->pos;
	}

	num_vis_balls = num_balls;
	pthread_mutex_unlock(&ballvis_mutex);
}

void Simulation::remove_dead_balls()
{
	list<SimBall*>::iterator iter = ball_list.begin();
	while(iter != ball_list.end()) {
		SimBall *ball = *iter;

		if(ball->pos.y > upper_limit || ball->pos.y < lower_limit) {
			if(bounds_callback) {
				bounds_callback(ball, bounds_callback_cls);
			}
			iter = ball_list.erase(iter);
			num_balls--;
		} else {
			iter++;
		}
	}
}

void Simulation::draw_static_objects(const henge::color &col) const
{
	for(size_t i=0; i<sobjects.size(); i++) {
		if(sobjects[i]->is_active()) {
			sobjects[i]->draw(col);
		}
	}
}
