#include <float.h>
#include "simobj.h"

#ifdef USE_GXBASE
#include "gxglut.h"
#else

#if defined(__APPLE__) && defined(__MACH__)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#endif	// USE_GXBASE


SimBall::SimBall(const Vector3 &pos, double rad, double mass, double rest)
{
	this->pos = pos;
	this->rad = rad;
	this->mass = mass;
	this->rest =rest;
	sim = true;
}

#define SWAP(type, a, b)	\
	do {					\
		type tmp = a;		\
		a = b;				\
		b = tmp;			\
	} while(0)

/* this dynamic sphere-sphere intersection does a ray-sphere intersection between
 * the relative trajectory of the first ball in respect to the second ball, and
 * extending the radius of the second ball to be its radius + the first ball's radius.
 */
bool ball_intersect(const SimBall *ball1, const SimBall *ball2, Contact *contact)
{
	// ray origin and direction (relative trajectory)
	Vector3 orig = ball1->pos;
	Vector3 dir = (ball1->end - ball1->pos) - (ball2->end - ball2->pos);

	// center/radius of the sphere we're intersecting the ray with
	Vector3 cent = ball2->pos;
	double rad = ball2->rad + ball1->rad;

	// if the spheres are moving away from each other, ignore intersections
	if(dot_product(ball1->pos - ball2->pos, dir) > 0.0) {
		return false;
	}

	double a = dot_product(dir, dir);
	if(a < XSMALL_NUMBER) {
		return false;
	}

	double b = 2.0 * dir.x * (orig.x - cent.x) + 2.0 * dir.y * (orig.y - cent.y) +
		2.0 * dir.z * (orig.z - cent.z);
	double c = dot_product(cent, cent) + dot_product(orig, orig) +
		2.0 * dot_product(orig, -cent) - SQ(rad);

	// solve the quadratic
	double d = SQ(b) - 4.0 * a * c;
	if(d < 0.0) {
		return false;
	}

	double sqrt_d = sqrt(d);
	double t1 = (-b + sqrt_d) / (2.0 * a);
	double t2 = (-b - sqrt_d) / (2.0 * a);

	if(t1 < XSMALL_NUMBER) t1 = t2;
	if(t2 < XSMALL_NUMBER) t2 = t1;

	float t = t1 < t2 ? t1 : t2;

	if(t < ERROR_MARGIN || t > 1.0) {
		return false;
	}

	contact->t = t;
	contact->pos = orig + dir * contact->t;
	contact->normal = (contact->pos - cent) / rad;
	contact->ball1 = (SimBall*)ball1;
	contact->ball2 = (SimBall*)ball2;
	contact->sobj = 0;
	return true;
}

bool ball_contact(const SimBall *ball1, const SimBall *ball2, Contact *contact)
{
	Vector3 normal = ball1->pos - ball2->pos;
	double dist = normal.length();
	double rad_sum = ball1->rad + ball2->rad;

	if(dist < rad_sum + ERROR_MARGIN) {
		contact->ball1 = (SimBall*)ball1;
		contact->ball2 = (SimBall*)ball2;
		contact->sobj = 0;
		contact->t = 0.0;
		contact->pos = ball2->pos + normal * 0.5;
		contact->normal = normal / dist;
		return true;
	}
	return false;
}

SimStaticObj::SimStaticObj()
{
	collision_func = 0;
	col_func_cls = 0;
	active = true;

	aabox.valid = false;
}

SimStaticObj::~SimStaticObj() {}

void SimStaticObj::set_collision_callback(bool (*func)(const SimStaticObj*,
			const SimBall*, const Contact*, void*), void *cls)
{
	collision_func = func;
	col_func_cls = cls;
}

bool SimStaticObj::signal_collision(const SimBall *ball, const Contact *cont) const
{
	if(collision_func) {
		return collision_func(this, ball, cont, col_func_cls);
	}
	return true;
}

void SimStaticObj::activate()
{
	active = true;
}

void SimStaticObj::deactivate()
{
	active = false;
}

/*
bool SimStaticObj::is_active() const
{
	return active;
}
*/

const AABox &SimStaticObj::get_bounds() const
{
	if(!aabox.valid) {
		((SimStaticObj*)this)->calc_bounds();
		aabox.valid = true;
	}
	return aabox;
}

SimQuad::SimQuad() {}

SimQuad::SimQuad(const Vector3 &a, const Vector3 &b, const Vector3 &c, const Vector3 &d)
{
	v[0] = a;
	v[1] = b;
	v[2] = c;
	v[3] = d;
	normal = cross_product(b - a, c - a).normalized();
}	

SimQuad::~SimQuad() {}

void SimQuad::calc_bounds()
{
	aabox.min = Vector3(FLT_MAX, FLT_MAX, FLT_MAX);
	aabox.max = Vector3(FLT_MIN, FLT_MIN, FLT_MIN);

	for(int i=0; i<4; i++) {
		if(v[i].x < aabox.min.x) aabox.min.x = v[i].x;
		if(v[i].y < aabox.min.y) aabox.min.y = v[i].y;
		if(v[i].z < aabox.min.z) aabox.min.z = v[i].z;
		if(v[i].x > aabox.max.x) aabox.max.x = v[i].x;
		if(v[i].y > aabox.max.y) aabox.max.y = v[i].y;
		if(v[i].z > aabox.max.z) aabox.max.z = v[i].z;
	}
}

bool SimQuad::intersect(const SimBall *ball, Contact *contact) const
{
	Vector3 dir = ball->end - ball->pos;

	double n_dot_dir = dot_product(normal, dir);
	if(n_dot_dir > -ERROR_MARGIN) {
		return false;	// ignore intersections from inside the object
	}

	Vector3 raised[4];

	/* we raise the vertices of the quad along its normal direction by the
	 * ammount of ball radius, so that the ball sweep volume/polygon
	 * intersection test becomes a simple ray/polygon test.
	 */
	raised[0] = v[0] + ball->rad * normal;

	Vector3 vo_vec = ball->pos - raised[0];
	double t = -dot_product(normal, vo_vec) / n_dot_dir;
	if(t < -ERROR_MARGIN || t > 1.0) {
		return false;
	}

	// calculate intersection (contact) point
	Vector3 pt = ball->pos + dir * t;

	// raise the rest of the vertices and test containment
	int sign_sum = 0;
	for(int i=0; i<4; i++) {
		if(i < 3) {
			raised[i + 1] = v[i + 1] + ball->rad * normal;
		}
		Vector3 cross = cross_product(raised[(i + 1) % 4] - raised[i], pt - raised[i]);
		sign_sum += dot_product(cross, normal) >= 0.0 ? 1 : -1;
	}

	if(abs(sign_sum) != 4) {
		return false;	// containment test failed
	}

	contact->t = t;
	contact->pos = pt;
	contact->normal = normal;
	contact->ball1 = (SimBall*)ball;
	contact->ball2 = 0;
	contact->sobj = (SimStaticObj*)this;
	return signal_collision(ball, contact);
}

bool SimQuad::contact(const SimBall *ball, Contact *contact) const
{
	Vector3 vo_vec = ball->pos - v[0];
	double dist = dot_product(vo_vec, normal);

	if(fabs(dist) > ball->rad + SMALL_NUMBER) {
		return false;
	}

	// project onto the plane and test for containment
	Vector3 proj = ball->pos - normal * dist;
	int sign_sum = 0;
	for(int i=0; i<4; i++) {
		Vector3 cross = cross_product(v[(i + 1) % 4] - v[i], proj - v[i]);
		sign_sum += dot_product(cross, normal) >= 0.0 ? 1 : -1;
	}

	if(abs(sign_sum) != 4) {
		return false;	// containment test failed
	}
	
	contact->ball1 = (SimBall*)ball;
	contact->ball2 = 0;
	contact->sobj = (SimStaticObj*)this;
	contact->pos = proj;
	contact->normal = normal;
	contact->t = 0.0;
	return signal_collision(ball, contact);
}

void SimQuad::draw(const henge::color &col) const
{
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);

	if(col.w < 1.0) {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDepthMask(GL_FALSE);
	}

	glPolygonOffset(-1.0, -1.0);
	glEnable(GL_POLYGON_OFFSET_FILL);

	glBegin(GL_QUADS);
	glColor4f(col.x, col.y, col.z, col.w);
	for(int i=0; i<4; i++) {
		glVertex3f(v[i].x, v[i].y, v[i].z);
	}
	glEnd();

	glDisable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(0.0, 0.0);
	glDepthMask(GL_TRUE);

	glDisable(GL_DEPTH_TEST);

	glBegin(GL_LINES);
	glColor3f(0, 1, 0);
	Vector3 start = (v[0] + v[1] + v[2] + v[3]) / 4.0;
	glVertex3f(start.x, start.y, start.z);
	glVertex3f(start.x + normal.x * 2, start.y + normal.y * 2, start.z + normal.z * 2);
	glEnd();

	glPopAttrib();
}


bool SimTrigger::contact(const SimBall *ball, Contact *contact) const
{
	return false;
}


SimCylinder::SimCylinder()
{
	rad = 1.0;
	min_z = -100.0;
	max_z = 100.0;
}

SimCylinder::SimCylinder(Vector2 pos, double rad)
{
	this->pos = pos;
	this->rad = rad;
	min_z = -100.0;
	max_z = 100.0;
}

SimCylinder::~SimCylinder() {}

void SimCylinder::calc_bounds()
{
	aabox.min = Vector3(pos.x - rad, pos.y - rad, min_z);
	aabox.max = Vector3(pos.x + rad, pos.y + rad, max_z);
}

bool SimCylinder::intersect(const SimBall *ball, Contact *contact) const
{
	// ray origin and direction (ball trajectory)
	Vector3 orig = ball->pos;
	Vector3 dir = ball->end - ball->pos;
	
	// center/radius of the sphere we're intersecting the ray with
	Vector2 cent = pos;
	double rad = this->rad + ball->rad;

	Vector2 cyl_to_ball = Vector2(ball->pos.x, ball->pos.y) - cent;
	Vector2 dir2 = Vector2(dir.x, dir.y);
	if(dot_product(dir2, cyl_to_ball) > 0.0) {
		return false;
	}

	double a = SQ(dir.x) + SQ(dir.y);
	if(a < XSMALL_NUMBER) {
		return false;
	}

	double b = 2.0 * dir.x * (orig.x - cent.x) + 2.0 * dir.y * (orig.y - cent.y);
	double c = SQ(cent.x) + SQ(cent.y) + SQ(orig.x) + SQ(orig.y) +
		2.0 * (-cent.x * orig.x - cent.y * orig.y) - SQ(rad);

	// solve the quadratic
	double d = SQ(b) - 4.0 * a * c;
	if(d < 0.0) {
		return false;
	}

	double sqrt_d = sqrt(d);
	double t1 = (-b + sqrt_d) / (2.0 * a);
	double t2 = (-b - sqrt_d) / (2.0 * a);

	if(t1 < XSMALL_NUMBER) t1 = t2;
	if(t2 < XSMALL_NUMBER) t2 = t1;

	float t = t1 < t2 ? t1 : t2;

	if(t < ERROR_MARGIN || t > 1.0) {
		return false;
	}

	contact->t = t;
	contact->pos = orig + dir * contact->t;
	contact->normal = (Vector2(contact->pos.x, contact->pos.y) - cent);
	contact->normal.normalize();
	contact->ball1 = (SimBall*)ball;
	contact->ball2 = 0;
	contact->sobj = (SimStaticObj*)this;
	return signal_collision(ball, contact);
}

bool SimCylinder::contact(const SimBall *ball, Contact *contact) const
{
	return false;
}

void SimCylinder::draw(const henge::color &col) const
{
	const AABox &box = get_bounds();

	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);

	if(col.w < 1.0) {
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDepthMask(GL_FALSE);
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(pos.x, pos.y, 0.0);
	glScalef((box.max.x - box.min.x) * 0.5, (box.max.y - box.min.y) * 0.5, box.max.z - box.min.z);

	glColor4f(col.x, col.y, col.z, col.w);
	glutSolidCube(1.0);

	glDepthMask(GL_TRUE);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glPopAttrib();
}
